/*
 *        spID3.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spString.h>
#include <sp/spKanji.h>

#include <sp/spChunkP.h>
#include <sp/spID3.h>

static char sp_id3_type_id_version_table[][3][5] = {
    {SP_ID3_FRAME_TYPE_ALBUM_V2_3, SP_ID3_FRAME_TYPE_ALBUM_V2_3, SP_ID3_FRAME_TYPE_ALBUM_V2_2},
    {SP_ID3_FRAME_TYPE_COMPOSER_V2_3, SP_ID3_FRAME_TYPE_COMPOSER_V2_3, SP_ID3_FRAME_TYPE_COMPOSER_V2_2},
    {SP_ID3_FRAME_TYPE_LYRICIST_V2_3, SP_ID3_FRAME_TYPE_LYRICIST_V2_3, SP_ID3_FRAME_TYPE_LYRICIST_V2_2},
    {SP_ID3_FRAME_TYPE_COPYRIGHT_V2_3, SP_ID3_FRAME_TYPE_COPYRIGHT_V2_3, SP_ID3_FRAME_TYPE_COPYRIGHT_V2_2},
    {SP_ID3_FRAME_TYPE_TRACK_V2_3, SP_ID3_FRAME_TYPE_TRACK_V2_3, SP_ID3_FRAME_TYPE_TRACK_V2_2},
    {SP_ID3_FRAME_TYPE_DISC_V2_3, SP_ID3_FRAME_TYPE_DISC_V2_3, SP_ID3_FRAME_TYPE_DISC_V2_2},
    {SP_ID3_FRAME_TYPE_TITLE_V2_3, SP_ID3_FRAME_TYPE_TITLE_V2_3, SP_ID3_FRAME_TYPE_TITLE_V2_2},
    {SP_ID3_FRAME_TYPE_ARTIST_V2_3, SP_ID3_FRAME_TYPE_ARTIST_V2_3, SP_ID3_FRAME_TYPE_ARTIST_V2_2},
    {SP_ID3_FRAME_TYPE_GENRE_V2_3, SP_ID3_FRAME_TYPE_GENRE_V2_3, SP_ID3_FRAME_TYPE_GENRE_V2_2},
    {SP_ID3_FRAME_TYPE_SOFTWARE_V2_3, SP_ID3_FRAME_TYPE_SOFTWARE_V2_3, SP_ID3_FRAME_TYPE_SOFTWARE_V2_2},
    {SP_ID3_FRAME_TYPE_TEMPO_V2_3, SP_ID3_FRAME_TYPE_TEMPO_V2_3, SP_ID3_FRAME_TYPE_TEMPO_V2_2},
    {SP_ID3_FRAME_TYPE_RELEASE_V2_4, SP_ID3_FRAME_TYPE_RELEASE_V2_3, SP_ID3_FRAME_TYPE_RELEASE_V2_2},
    {SP_ID3_FRAME_TYPE_ISRC_V2_3, SP_ID3_FRAME_TYPE_ISRC_V2_3, ""},
    {SP_ID3_FRAME_TYPE_PUBLISHER_V2_3, SP_ID3_FRAME_TYPE_PUBLISHER_V2_3, SP_ID3_FRAME_TYPE_PUBLISHER_V2_2},
    {"", "", ""},
};

static const char *spGetID3CurrentVersionTypeId(int version, const char *type)
{
    int i;
    int version_index;

    if (version >= 4) {
        version_index = 0;
    } else if (version == 3) {
        version_index = 1;
    } else {
        version_index = 2;
    }

    for (i = 0;; i++) {
        if (sp_id3_type_id_version_table[i][0][0] == NUL) {
            break;
        }
        spDebug(100, "spGetID3CurrentVersionTypeId", "i = %d: type = %s, entry = %s\n",
                i, type, sp_id3_type_id_version_table[i][version_index]);

        if (streq(sp_id3_type_id_version_table[i][0], type)
            || streq(sp_id3_type_id_version_table[i][1], type)
            || streq(sp_id3_type_id_version_table[i][2], type)) {
            if (sp_id3_type_id_version_table[i][version_index][0] == NUL) {
                return NULL;
            } else {
                return sp_id3_type_id_version_table[i][version_index];
            }
        }
    }

    return NULL;
}

const char *spGetID3HeaderCurrentVersionTypeId(spID3Header *header, const char *type)
{
    return spGetID3CurrentVersionTypeId(header->version, type);
}

spID3Header *spGetID3HeaderFromFrame(spID3Frame *frame)
{
    spID3Frame *parent;

    parent = frame;
    while (parent->parent != NULL) {
        parent = parent->parent;
    }

    return (spID3Header *)parent;
}

spBool spIsID3NullFrameHeader(spID3FrameHeader *frame_header)
{
    if (frame_header->type[0] == 0 && frame_header->type[1] == 0 && frame_header->type[2] == 0) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spIsID3FrameUnsynchronized(spID3Frame *frame)
{
    spID3Header *header;

    header = spGetID3HeaderFromFrame(frame);

    if (header->version >= 4) {
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_UNSYNCHRONISATION_MASK) {
            return SP_TRUE;
        } else {
            return SP_FALSE;
        }
    } else {
        if (header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK) {
            return SP_TRUE;
        } else {
            return SP_FALSE;
        }
    }
}

void spSetID3FrameUnsynchronizedFlag(spID3Frame *frame)
{
    spID3Header *header;

    header = spGetID3HeaderFromFrame(frame);

    if (header->version >= 4) {
        frame->header.flag2 |= SP_ID3_FRAME_HEADER_FLAG2_V2_4_UNSYNCHRONISATION_MASK;
    } else {
        header->header.flag1 |= SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK;
    }

    return;
}

void spUpdateID3FrameUnsynchronizedFlag(spID3Frame *frame)
{
    spID3Header *header;

    header = spGetID3HeaderFromFrame(frame);

    if (header->unsync_count > 0) {
        if (header->version >= 4) {
            frame->header.flag2 |= SP_ID3_FRAME_HEADER_FLAG2_V2_4_UNSYNCHRONISATION_MASK;
        } else {
            header->header.flag1 |= SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK;
        }
    }

    return;
}

long spReadID3SyncSafeLong32(unsigned long *value, FILE *fp)
{
    long nread;
    unsigned char buf[4];

    if ((nread = (long)fread(buf, 1, 4, fp)) != 4) {
        return nread;
    }

    *value = (0x7f & (unsigned long)buf[0]) << 21 | (0x7f & (unsigned long)buf[1]) << 14
        | (0x7f & (unsigned long)buf[2]) << 7 | (0x7f & (unsigned long)buf[3]);

    return nread;
}

long spWriteID3SyncSafeLong32(unsigned long value, FILE *fp)
{
    long nwrite;
    unsigned char buf[4];

    buf[0] = (unsigned char)(0x7f & (value >> 21));
    buf[1] = (unsigned char)(0x7f & (value >> 14));
    buf[2] = (unsigned char)(0x7f & (value >> 7));
    buf[3] = (unsigned char)(0x7f & value);

    if ((nwrite = (long)fwrite(buf, 1, 4, fp)) != 4) {
        return nwrite;
    }

    return nwrite;
}

long spReadID3UnsynchronizedBuffer(unsigned char *buf, long need_len, long read_len, long block_len, long stride,
                                   int swap_flag, int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    long k, n;
    long ret;
    long nread;
    unsigned char prev_value;
    long shift;
    long swap_size = 0;

    prev_value = 0;
    if (last_value != NULL) prev_value = *last_value;

    if (swap_flag) {
        swap_size = block_len;
    }
    
    if (swap_size > 1) {
        shift = swap_size - 1;
    } else {
        shift = 0;
    }

    for (k = 0, nread = 0;;) {
        if (!SP_AIFF_NEED_SWAP && block_len < stride) {
            if (k % stride == 0) {
                n = k;
                k += (stride - block_len);
                for (; n < k; n++) {
                    buf[k] = 0;
                }
            }
        }

        if ((ret = (long)fread(&buf[k + shift], 1, 1, fp)) != 1) {
            spDebug(10, "spReadID3UnsynchronizedBuffer", "fread failed: %ld\n", ret);
            return ret;
        }
        ++nread;

        if (unsync_flag) {
            if (prev_value == 0xff && buf[k + shift] == 0) {
                if ((ret = (long)fread(&buf[k + shift], 1, 1, fp)) != 1) {
                    spDebug(10, "spReadID3UnsynchronizedBuffer", "fread failed: %ld\n", ret);
                    return ret;
                }
                if (unsync_count != NULL) {
                    *unsync_count += 1;
                }
                spDebug(10, "spReadID3UnsynchronizedBuffer", "updated buf[%ld] = %x, nread = %ld\n", k, buf[k], nread);
                ++nread;
            }
        }

        prev_value = buf[k + shift];
        if (last_value != NULL) *last_value = prev_value;
        
        ++k;
        
        if (SP_AIFF_NEED_SWAP && block_len < stride) {
            if (k % stride == block_len) {
                n = k;
                k += (stride - block_len);
                for (; n < k; n++) {
                    buf[k] = 0;
                }
            }
        }
        
        if (read_len > 0 && nread >= read_len) {
            break;
        }
        
        if (need_len > 0 && k >= need_len) {
            break;
        }
        
        if (swap_size > 1) {
            shift -= 2;
            if (shift <= -swap_size) {
                shift = swap_size - 1;
            }
        }
    }

    return nread;
}

long spReadID3UnsynchronizedULong32(unsigned long *value, int unsync_flag,
                                    unsigned char *last_value, long *unsync_count, FILE *fp)
{
    long nread;
    
    if ((nread = spReadID3UnsynchronizedBuffer((unsigned char *)value, 4, 0, 4, 4, SP_AIFF_NEED_SWAP,
                                               unsync_flag, last_value, unsync_count, fp)) <= 0) {
        spDebug(10, "spReadID3UnsynchronizedULong32", "spReadID3UnsynchronizedBuffer failed\n");
        return 0;
    }

    return nread;
}

long spWriteID3UnsynchronizedBuffer(unsigned char *buf, long buf_len, long block_len, long stride,
                                    int swap_flag, int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    long k;
    long ret;
    long nwrite;
    unsigned char prev_value;
    unsigned char c, zc;
    long shift;
    long swap_size = 0;

    zc = 0;
    prev_value = 0;
    if (last_value != NULL) prev_value = *last_value;

    if (swap_flag) {
        swap_size = block_len;
    }
    
    if (swap_size > 1) {
        shift = swap_size - 1;
    } else {
        shift = 0;
    }

    for (k = 0, nwrite = 0;;) {
        if (!SP_AIFF_NEED_SWAP && block_len < stride) {
            if (k % stride == 0) k += (stride - block_len);
        }

        c = buf[k + shift];

        if (unsync_flag) {
            if (prev_value == 0xff && ((c & 0xe0) == 0xe0 || c == 0)) { /* 0xe0 = 11100000 */
                zc = 0;
                if ((ret = (long)fwrite(&zc, 1, 1, fp)) != 1) {
                    spDebug(10, "spWriteID3UnsynchronizedBuffer", "fwrite failed: %ld\n", ret);
                    return ret;
                }
                if (last_value != NULL) *last_value = zc;
                if (unsync_count != NULL) {
                    *unsync_count += 1;
                }
                spDebug(10, "spWriteID3UnsynchronizedBuffer", "unsync: %ld\n", nwrite);
                ++nwrite;
            }
        }

        if ((ret = (long)fwrite(&c, 1, 1, fp)) != 1) {
            spDebug(10, "spWriteID3UnsynchronizedBuffer", "fwrite failed: %ld\n", ret);
            return ret;
        }
        if (last_value != NULL) *last_value = c;
        ++nwrite;

        ++k;

        if (SP_AIFF_NEED_SWAP && block_len < stride) {
            if (k % stride == block_len) k += (stride - block_len);
        }

        if (k >= buf_len) {
            break;
        }
        
        if (swap_size > 1) {
            shift -= 2;
            if (shift <= -swap_size) {
                shift = swap_size - 1;
            }
        }
        
        prev_value = c;
    }

    return nwrite;
}

long spWriteID3UnsynchronizedULong32(unsigned long value, int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    long nwrite;

    if ((nwrite = spWriteID3UnsynchronizedBuffer((unsigned char *)&value, 4, 4, 4, SP_AIFF_NEED_SWAP, unsync_flag,
                                                 last_value, unsync_count, fp)) <= 0) {
        spDebug(10, "spWriteID3UnsynchronizedULong32", "spWriteID3UnsynchronizedBuffer failed\n");
        return 0;
    }

    return nwrite;
}

spID3FrameSize spGetID3FrameContentSizeForChunkFileSpec(spID3Frame *frame)
{
    spDebug(100, "spGetID3FrameContentSizeForChunkFileSpec", "frame->header.size = %lu\n", (unsigned long)frame->header.size);
    return frame->header.size;
}

spID3FrameSize spGetID3FrameContentSize(spID3Frame *frame)
{
    if (frame == NULL) return 0;
    
    return spGetID3FrameContentSizeForChunkFileSpec(frame);
}

spBool spSetID3FrameContentSizeForChunkFileSpec(spID3Frame *frame, spID3FrameSize size)
{
    frame->header.size = (unsigned long)size;
    spDebug(100, "spSetID3FrameContentSizeForChunkFileSpec", "frame->header.size = %lu\n", (unsigned long)frame->header.size);

    return SP_TRUE;
}

spID3FrameSize spGetID3FrameHeaderExtraSize(spID3Frame *frame)
{
    spID3Header *header;
    spID3FrameSize size = 0;
    
    header = spGetID3HeaderFromFrame(frame);
    
    if (header->version == 3) {
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_COMPRESSION_MASK) {
            size += 4;
        }
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_ENCRIPTION_MASK) {
            size += 1;
        }
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_GROUPING_MASK) {
            size += 1;
        }
    } else if (header->version == 4) {
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_GROUPING_MASK) {
            size += 1;
        }
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_COMPRESSION_MASK) {
            size += 4;
        }
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_ENCRIPTION_MASK) {
            size += 1;
        }
        if (frame->header.flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_DATA_LENGTH_MASK) {
            size += 4;
        }
    }
        
    return size;
}

spID3FrameSize spGetID3FrameHeaderSize(spID3Frame *frame)
{
    spID3Header *header;

    header = spGetID3HeaderFromFrame(frame);
    
    if (header->version == 2) {
        return 6;
    } else {
        return 10 + spGetID3FrameHeaderExtraSize(frame);
    }
}

spID3FrameSize spGetID3FrameSize(spID3Frame *frame, char *string)
{
    if (string != NULL) {
        memcpy(string, frame->header.type, 4);
        string[4] = NUL;
    }

    return frame->header.size + spGetID3FrameHeaderSize(frame);
}

spID3FrameSize spGetID3HeaderSize(spID3Header *header)
{
    return spGetID3FrameSize((spID3Frame *)header, NULL);
}

spID3FrameSize spReadID3FrameHeader(spID3Frame *parent, spID3FrameHeader *frame_header, int swap, FILE *fp)
{
    long len;
    spID3FrameSize nread;
    spID3Header *header;
    unsigned char last_value = 0;

    header = spGetID3HeaderFromFrame(parent);

    memset(frame_header, 0, sizeof(spID3FrameHeader));

    if (header->version == 2) {
        len = 3;
        frame_header->type[3] = ' ';
    } else {
        len = 4;
    }
    
    if (fread(frame_header->type, 1, len, fp) != len) {
        spDebug(10, "spReadID3FrameHeader", "Can't read frame type.\n");
        return 0;
    }
    nread = len;
    spDebug(10, "spReadID3FrameHeader", "len = %d, type = %c%c%c%c\n",
            len, frame_header->type[0], frame_header->type[1], frame_header->type[2], frame_header->type[3]);
    
    if (header->version == 2) {
        if ((len = spReadID3UnsynchronizedBuffer((unsigned char *)&frame_header->size, 3, 0, 3, 4, SP_AIFF_NEED_SWAP,
                                                 header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK,
                                                 &last_value, &header->unsync_count, fp)) <= 0) {
            spDebug(10, "spReadID3FrameHeader", "spReadID3UnsynchronizedBuffer failed\n");
            return 0;
        }
        nread += 3;
    } else {
        if (header->version == 3) {
            if ((len = spReadID3UnsynchronizedULong32(&frame_header->size,
                                                      header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK,
                                                      &last_value, &header->unsync_count, fp)) <= 0) {
                spDebug(10, "spReadID3FrameHeader", "spReadID3UnsynchronizedULong32 failed\n");
                return 0;
            }
        } else {
            if (spReadID3SyncSafeLong32(&frame_header->size, fp) != 4) {
                spDebug(10, "spReadID3Header", "Can't read size: %ld\n", (long)nread);
                return 0;
            }
        }
        nread += 4;
        spDebug(10, "spReadID3FrameHeader", "size = %ld\n", frame_header->size);

        if (fread(&frame_header->flag1, 1, 1, fp) != 1
            || fread(&frame_header->flag2, 1, 1, fp) != 1) {
            return 0;
        }
        nread += 2;
        spDebug(10, "spReadID3FrameHeader", "flag1 = %lx, flag2 = %lx\n", frame_header->flag1, frame_header->flag2);
        
        if (header->version == 3) {
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_COMPRESSION_MASK) {
                if ((len = spReadID3UnsynchronizedULong32(&frame_header->extra.v2_3.decompressed_size,
                                                          header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK,
                                                          &last_value, &header->unsync_count, fp)) <= 0) {
                    spDebug(10, "spReadID3FrameHeader", "spReadID3UnsynchronizedULong32 failed\n");
                    return 0;
                }
                nread += 4;
                spDebug(10, "spReadID3FrameHeader", "decompressed_size = %ld\n", frame_header->extra.v2_3.decompressed_size);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_ENCRIPTION_MASK) {
                if (fread(&frame_header->extra.v2_3.encryption_method, 1, 1, fp) != 1) {
                    return 0;
                }
                nread += 1;
                spDebug(10, "spReadID3FrameHeader", "encryption_mehtod = %d\n", frame_header->extra.v2_3.encryption_method);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_GROUPING_MASK) {
                if (fread(&frame_header->extra.v2_3.group_id, 1, 1, fp) != 1) {
                    return 0;
                }
                nread += 1;
                spDebug(10, "spReadID3FrameHeader", "group_id = %d\n", frame_header->extra.v2_3.group_id);
            }
        } else if (header->version == 4) {
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_GROUPING_MASK) {
                if (fread(&frame_header->extra.v2_4.group_id, 1, 1, fp) != 1) {
                    return 0;
                }
                nread += 1;
                spDebug(10, "spReadID3FrameHeader", "group_id = %d\n", frame_header->extra.v2_4.group_id);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_ENCRIPTION_MASK) {
                if (fread(&frame_header->extra.v2_4.encryption_method, 1, 1, fp) != 1) {
                    return 0;
                }
                nread += 1;
                spDebug(10, "spReadID3FrameHeader", "encryption_mehtod = %d\n", frame_header->extra.v2_4.encryption_method);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_COMPRESSION_MASK) {
                if (spReadID3SyncSafeLong32(&frame_header->extra.v2_4.data_length_indicator, fp) != 4) {
                    spDebug(10, "spReadID3Header", "Can't read size: %ld\n", (long)nread);
                    return 0;
                }
                nread += 4;
                spDebug(10, "spReadID3FrameHeader", "decompressed_size = %ld\n", frame_header->extra.v2_4.data_length_indicator);
            }
        }
    }
    spDebug(10, "spReadID3FrameHeader", "size = %ld\n", frame_header->size);
    
    return nread;
}

spID3FrameSize spWriteID3FrameHeader(spID3Frame *frame, spID3FrameHeader *frame_header, int swap, FILE *fp)
{
    long len;
    spID3FrameSize nwrite;
    spID3Header *header;
    unsigned long size;
    unsigned char last_value = 0;

    header = spGetID3HeaderFromFrame(frame);

    if (header->version == 2) {
        len = 3;
    } else {
        len = 4;
    }
    
    spDebug(10, "spWriteID3FrameHeader", "len = %d, type = %c%c%c%c\n",
            len, frame_header->type[0], frame_header->type[1], frame_header->type[2], frame_header->type[3]);
    if (fwrite(frame_header->type, 1, len, fp) != len) {
        spDebug(10, "spWriteID3FrameHeader", "Can't write frame type.\n");
        return 0;
    }
    nwrite = len;

    if (spIsID3NullFrameHeader(frame_header)) {
        size = 0;
    } else {
        size = frame_header->size;
    }
    
    spDebug(10, "spWriteID3FrameHeader", "size = %ld\n", size);
    
    if (header->version == 2) {
        if ((len = spWriteID3UnsynchronizedBuffer((unsigned char *)&size, 3, 3, 4, SP_AIFF_NEED_SWAP,
                                                  header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK,
                                                  &last_value, &header->unsync_count, fp)) <= 0) {
            spDebug(10, "spWriteID3FrameHeader", "spWriteID3UnsynchronizedBuffer failed\n");
            return 0;
        }
        nwrite += 3;
    } else {
        if (header->version == 3) {
            if ((len = spWriteID3UnsynchronizedULong32(size,
                                                       header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK,
                                                       &last_value, &header->unsync_count, fp)) <= 0) {
                spDebug(10, "spWriteID3FrameHeader", "spWriteID3UnsynchronizedULong32 failed\n");
                return 0;
            }
        } else {
            if (spWriteID3SyncSafeLong32(size, fp) != 4) {
                spDebug(10, "spWriteID3Header", "Can't write size: %ld\n", (long)nwrite);
                return 0;
            }
        }
        nwrite += 4;

        spDebug(10, "spWriteID3FrameHeader", "flag1 = %lx, flag2 = %lx\n", frame_header->flag1, frame_header->flag2);
        if (fwrite(&frame_header->flag1, 1, 1, fp) != 1
            || fwrite(&frame_header->flag2, 1, 1, fp) != 1) {
            return 0;
        }
        nwrite += 2;
        
        if (header->version == 3) {
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_COMPRESSION_MASK) {
                if ((len = spWriteID3UnsynchronizedULong32(frame_header->extra.v2_3.decompressed_size,
                                                           header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK,
                                                           &last_value, &header->unsync_count, fp)) <= 0) {
                    spDebug(10, "spWriteID3FrameHeader", "spWriteID3UnsynchronizedULong32 failed\n");
                    return 0;
                }
                nwrite += 4;
                spDebug(10, "spWriteID3FrameHeader", "decompressed_size = %ld\n", frame_header->extra.v2_3.decompressed_size);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_ENCRIPTION_MASK) {
                if (fwrite(&frame_header->extra.v2_3.encryption_method, 1, 1, fp) != 1) {
                    return 0;
                }
                nwrite += 1;
                spDebug(10, "spWriteID3FrameHeader", "encryption_mehtod = %d\n", frame_header->extra.v2_3.encryption_method);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_3_GROUPING_MASK) {
                if (fwrite(&frame_header->extra.v2_3.group_id, 1, 1, fp) != 1) {
                    return 0;
                }
                nwrite += 1;
                spDebug(10, "spWriteID3FrameHeader", "group_id = %d\n", frame_header->extra.v2_3.group_id);
            }
        } else if (header->version == 4) {
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_GROUPING_MASK) {
                if (fwrite(&frame_header->extra.v2_4.group_id, 1, 1, fp) != 1) {
                    return 0;
                }
                nwrite += 1;
                spDebug(10, "spWriteID3FrameHeader", "group_id = %d\n", frame_header->extra.v2_4.group_id);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_ENCRIPTION_MASK) {
                if (fwrite(&frame_header->extra.v2_4.encryption_method, 1, 1, fp) != 1) {
                    return 0;
                }
                nwrite += 1;
                spDebug(10, "spWriteID3FrameHeader", "encryption_mehtod = %d\n", frame_header->extra.v2_4.encryption_method);
            }
            if (frame_header->flag2 & SP_ID3_FRAME_HEADER_FLAG2_V2_4_COMPRESSION_MASK) {
                if (spWriteID3SyncSafeLong32(frame_header->extra.v2_4.data_length_indicator, fp) != 4) {
                    spDebug(10, "spWriteID3Header", "Can't write size: %ld\n", (long)nwrite);
                    return 0;
                }
                nwrite += 4;
                spDebug(10, "spWriteID3FrameHeader", "decompressed_size = %ld\n", frame_header->extra.v2_4.data_length_indicator);
            }
        }
    }
    
    spDebug(10, "spWriteID3FrameHeader", "done: size = %ld\n", frame_header->size);
    
    return nwrite;
}

spID3FrameSize spSkipID3Frame(spID3Frame *parent, spID3FrameHeader *frame_header, spID3FrameSize remain_size, int swap, FILE *fp)
{
    spID3FrameSize skiplen;
    spID3Header *header;

    header = spGetID3HeaderFromFrame(parent);
    
    spDebug(10, "spSkipID3Frame", "remain_size = %ld, frame_header->size = %ld, frame_header->type = %c%c%c%c\n",
            (long)remain_size, frame_header->size, frame_header->type[0], frame_header->type[1], frame_header->type[2], frame_header->type[3]);

    if (spIsID3NullFrameHeader(frame_header)) {
        skiplen = remain_size;
        spDebug(10, "spSkipID3Frame", "**** padding found ****, skiplen = %ld\n", (long)skiplen);
        if (spSeekFile(fp, skiplen, SEEK_CUR) != 0) {
            return 0;
        }
        frame_header->size = (unsigned long)skiplen;
    } else {
    
        if (frame_header->size <= 0) {
            skiplen = 0;
        } else {
            if (header->version == 4 || !(header->header.flag1 & SP_ID3_HEADER_FLAG_UNSYNCHRONISATION_MASK)) {
                skiplen = frame_header->size;
                if (spSeekFile(fp, skiplen, SEEK_CUR) != 0) {
                    return 0;
                }
            } else {
                unsigned char c, pc;
                long len;

                pc = 0;
                for (skiplen = 0;;) {
                    if ((len = spReadID3UnsynchronizedBuffer(&c, 1, 0, 1, 1, SP_FALSE, SP_TRUE, &pc, &header->unsync_count, fp)) <= 0) {
                        return 0;
                    }

                    if (++skiplen >= frame_header->size) {
                        break;
                    }
                    pc = c;
                }
            }
        }
    }

    return skiplen;
}

spID3FrameSize spWriteID3ContentRest(spID3Frame *frame, long depth, spID3FrameSize remain_size, spBool rewrite_flag, int swap, FILE *fp)
{
    spID3FrameSize len;
    spID3FrameSize nwrite;
    spID3FrameSize nremain;
    spID3FrameSize total_nwrite;
    static unsigned char buf[8] = {0, 0, 0, 0, 0, 0, 0, 0};

    spDebug(80, "spWriteID3ContentRest", "frame->header.size = %ld, frame->header.type = %c%c%c%c\n",
            frame->header.size, frame->header.type[0], frame->header.type[1], frame->header.type[2], frame->header.type[3]);
    
    spDebug(80, "spWriteID3ContentRest", "depth = %ld, remain_size %ld\n", depth, (long)remain_size);

    total_nwrite = 0;
    nremain = remain_size;
    
    spDebug(80, "spWriteID3ContentRest", "nremain = %ld\n", (long)nremain);
    
    while (nremain > 0) {
        len = MIN(8, nremain);
        spDebug(100, "spWriteID3ContentRest", "len = %ld, nremain = %ld\n", (long)len, (long)nremain);
        if ((nwrite = fwrite(buf, 1, (long)len, fp)) != len) {
            spDebug(10, "spWriteID3ContentRest", "fwrite zeros failed: nwrite = %ld\n", (long)nwrite);
            break;
        }
        spDebug(100, "spWriteID3ContentRest", "fwrite zeros done: nwrite = %ld\n", (long)nwrite);

        nremain -= nwrite;
        total_nwrite += nwrite;
    }

    spDebug(80, "spWriteID3ContentRest", "done: total_nwrite = %ld\n", (long)total_nwrite);
    
    return total_nwrite;
}

long spReadID3String(unsigned char encoding, unsigned char *buf, long need_len, long read_len, spBool *bom_flag,
                     spBool *swap_flag, int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    long len;
    long nread;
    long uc;

    uc = 0;
    if ((nread = spReadID3UnsynchronizedBuffer(buf, need_len, read_len, 1, 1, SP_FALSE, unsync_flag, last_value, &uc, fp)) <= 0) {
        spDebug(10, "spReadID3String", "spReadID3UnsynchronizedBuffer failed\n");
        return 0;
    }
    len = nread - uc;
    if (unsync_count != NULL) {
        *unsync_count += uc;
    }
    
    if (bom_flag != NULL) *bom_flag = SP_FALSE;
    
    spDebug(80, "spReadID3String", "encoding = %d\n", encoding);
    
    if (encoding == 0/* ISO-8859-1 */ || encoding == 3/* UTF-8 */) {
        if (swap_flag != NULL) *swap_flag = SP_FALSE;
    } else {
        int swap = 0;
        unsigned short *sbuf = (unsigned short *)buf;

        if (encoding != 2) {        /* with BOM */
            if (sbuf[0] == 0xfffe) {
                swap = 1;
            }
            
            if (sbuf[0] == 0xfffe || sbuf[0] == 0xfeff) {
                long k;

                spDebug(100, "spReadID3String", "BOM found\n");
                
                len -= 2;
                
                for (k = 0; k < len / 2; k++) {
                    sbuf[k] = sbuf[k + 1];
                }
                sbuf[k] = 0;

                if (bom_flag != NULL) *bom_flag = SP_TRUE;
            } else {
                swap = SP_AIFF_NEED_SWAP; /* no BOM */
            }
        } else {        /* 2: without BOM */
            swap = SP_AIFF_NEED_SWAP;
        }
        spDebug(100, "spReadID3String", "swap = %d\n", swap);

        if (swap) {
            spSwapShort((short *)sbuf, len / 2);
        }
        
        if (swap_flag != NULL) *swap_flag = swap;
    }

    spDebug(80, "spReadID3String", "done: nread = %ld\n", nread);
    
    return nread;
}

long spWriteID3String(unsigned char encoding, unsigned char *buf, long buf_len, spBool bom_flag,
                      spBool swap_flag, int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    long nwrite;
    long uc;
    long bom_size = 0;
    
    spDebug(80, "spWriteID3String", "encoding = %d, buf_len = %ld, bom_flag = %d, swap_flag = %d\n",
            encoding, buf_len, bom_flag, swap_flag);
    
    if (encoding == 0/* ISO-8859-1 */ || encoding == 3/* UTF-8 */) {
        /* do nothing */
    } else {
        if (bom_flag && encoding != 2) {        /* with BOM */
            short bom = 0xfeff;

            spDebug(80, "spWriteID3String", "write BOM\n");
            
            uc = 0;
            if ((nwrite = spWriteID3UnsynchronizedBuffer((unsigned char *)&bom, 2, 2, 2, swap_flag, unsync_flag, last_value, &uc, fp)) <= 0) {
                spDebug(10, "spWriteID3String", "Can't write BOM of UTF string.\n");
                return nwrite;
            }
            if (unsync_count != NULL) {
                *unsync_count += uc;
            }
            bom_size = 2;
        }
    }

    uc = 0;
    if ((nwrite = spWriteID3UnsynchronizedBuffer(buf, buf_len, 2, 2, swap_flag, unsync_flag, last_value, &uc, fp)) <= 0) {
        spDebug(10, "spWriteID3String", "spWriteID3UnsynchronizedBuffer failed\n");
    } else {
        if (unsync_count != NULL) {
            *unsync_count += uc;
        }
    }
    
    spDebug(80, "spWriteID3String", "done: nwrite = %ld\n", nwrite);
    
    if (nwrite > 0) {
        return nwrite + bom_size;
    } else {
        return nwrite;
    }
}

spID3FrameSize spReadID3DataFrame(spChunk *parent, long depth, void *frame, int swap, FILE *fp)
{
    long len;
    long need_len, read_len;
    long uc;
    unsigned char last_value = 0;
    spID3FrameSize nread;
    spID3Header *header;
    spID3DataFrame *dframe = (spID3DataFrame *)frame;

    spDebug(80, "spReadID3DataFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)parent);
    
    dframe->data = xspAlloc(dframe->header.size + 1, unsigned char);

    if (header->version == 4) {
        need_len = 0;
        read_len = dframe->header.size;
    } else {
        need_len = dframe->header.size;
        read_len = 0;
    }

    uc = 0;
    if ((len = spReadID3UnsynchronizedBuffer(dframe->data, need_len, read_len, 1, 1, SP_FALSE, SP_TRUE, &last_value, &uc, fp)) <= 0) {
        spDebug(10, "spReadID3DataFrame", "spReadID3UnsynchronizedBuffer failed\n");
        return 0;
    }
    dframe->data[len - uc] = 0;        /* just in case */
    header->unsync_count += uc;
    nread = len;
    
    spDebug(80, "spReadID3DataFrame", "done: nread = %ld / %ld\n", (long)nread, dframe->header.size);
    
    return nread;
}

spID3FrameSize spWriteID3DataFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long len;
    long uc;
    spID3FrameSize nwrite;
    spID3Header *header;
    spID3DataFrame *dframe = (spID3DataFrame *)frame;

    spDebug(80, "spWriteID3DataFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)frame);
    
    uc = 0;
    if ((len = spWriteID3UnsynchronizedBuffer(dframe->data, dframe->header.size, 1, 1, 0, 1, 0, &uc, fp)) <= 0) {
        spDebug(10, "spWriteID3DataFrame", "spWriteID3UnsynchronizedBuffer failed\n");
        return 0;
    }
    header->unsync_count += uc;
    spUpdateID3FrameUnsynchronizedFlag((spID3Frame *)dframe);
    nwrite = len;
    
    spDebug(80, "spWriteID3DataFrame", "done: nwrite = %ld / %ld\n", (long)nwrite, dframe->header.size);
    
    return nwrite;
}

spBool spFreeID3DataFrame(void *frame)
{
    spID3DataFrame *dframe = (spID3DataFrame *)frame;

    if (dframe->data != NULL) {
        xspFree(dframe->data);
        dframe->data = NULL;
    }
    
    return SP_TRUE;
}

spID3FrameSize spCopyID3DataFrame(spChunk *dest_frame, spChunk *src_frame, long depth)
{
    spID3DataFrame *dest_dframe = (spID3DataFrame *)dest_frame;
    spID3DataFrame *src_dframe = (spID3DataFrame *)src_frame;

    spDebug(80, "spCopyID3DataFrame", "in\n");

    dest_dframe->data = xspAlloc(src_dframe->header.size + 1, unsigned char);
    memcpy(dest_dframe->data, src_dframe->data, src_dframe->header.size);
    dest_dframe->data[src_dframe->header.size] = 0;        /* just in case */
    
    return src_dframe->header.size;
}

spID3FrameSize spReadID3TextFrame(spChunk *parent, long depth, void *frame, int swap, FILE *fp)
{
    long len, len_wo_bom;
    long need_len, read_len;
    long uc;
    unsigned char last_value = 0;
    spID3FrameSize nread;
    spID3Header *header;
    spID3TextFrame *tframe = (spID3TextFrame *)frame;

    spDebug(80, "spReadID3TextFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)parent);
    
    if (fread(&tframe->encoding, 1, 1, fp) != 1) {
        return 0;
    }
    nread = 1;
    spDebug(80, "spReadID3TextFrame", "encoding = %d\n", tframe->encoding);

    tframe->string = xspAlloc(tframe->header.size - 1 + 2, unsigned char);

    if (header->version == 4) {
        need_len = 0;
        read_len = tframe->header.size - 1;
    } else {
        need_len = tframe->header.size - 1;
        read_len = 0;
    }

    uc = 0;
    if ((len = spReadID3String(tframe->encoding, tframe->string, need_len, read_len, &tframe->bom_flag, &tframe->swap_flag,
                               spIsID3FrameUnsynchronized((spID3Frame *)tframe), &last_value, &uc, fp)) <= 0) {
        spDebug(10, "spReadID3TextFrame", "Can't read string.\n");
        return 0;
    }
    spDebug(80, "spReadID3TextFrame", "bom_flag = %d, swap_flag = %d\n", tframe->bom_flag, tframe->swap_flag);
    len_wo_bom = tframe->bom_flag ? len - 2 : len;
    tframe->string[len_wo_bom - uc] = 0;
    tframe->string[len_wo_bom - uc + 1] = 0;
    header->unsync_count += uc;
    nread += len;
    spDebug(80, "spReadID3TextFrame", "len = %ld, string = `%s'\n", len, tframe->string);
    
    spDebug(80, "spReadID3TextFrame", "done: nread = %ld / %ld\n", (long)nread, tframe->header.size);
    
    return nread;
}

spID3FrameSize spWriteID3TextFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long len;
    long write_len;
    long uc;
    unsigned char last_value = 0;
    spID3FrameSize nwrite;
    spID3Header *header;
    spID3TextFrame *tframe = (spID3TextFrame *)frame;

    spDebug(80, "spWriteID3TextFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)frame);
    
    spDebug(80, "spWriteID3TextFrame", "encoding = %d, bom_flag = %d\n", tframe->encoding, tframe->bom_flag);
    if (fwrite(&tframe->encoding, 1, 1, fp) != 1) {
        return 0;
    }
    nwrite = 1;

    write_len = tframe->header.size - 1;
    if (tframe->bom_flag) {
        write_len -= 2;
    }

    spDebug(80, "spWriteID3TextFrame", "write_len = %ld, string = `%s'\n", write_len, tframe->string);
    uc = 0;
    if ((len = spWriteID3String(tframe->encoding, tframe->string, write_len, tframe->bom_flag, tframe->swap_flag,
                                spIsID3FrameUnsynchronized((spID3Frame *)tframe), &last_value, &uc, fp)) <= 0) {
        spDebug(10, "spWriteID3TextFrame", "Can't write string.\n");
        return 0;
    }
    header->unsync_count += uc;
    spUpdateID3FrameUnsynchronizedFlag((spID3Frame *)tframe);
    nwrite += len;
    
    spDebug(80, "spWriteID3TextFrame", "done: nwrite = %ld / %ld, unsync_count = %ld\n",
            (long)nwrite, tframe->header.size, header->unsync_count);
    
    return nwrite;
}

static spID3FrameSize readID3LangTextFrame(spChunk *parent, long depth, void *frame, int swap, FILE *fp, spBool lang_flag)
{
    long len;
    long need_len, read_len;
    long uc;
    unsigned char last_value = 0;
    spID3FrameSize nread;
    spID3Header *header;
    spID3LangTextFrame *ltframe = (spID3LangTextFrame *)frame;

    spDebug(80, "readID3LangTextFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)parent);
    
    ltframe->string = NULL;
    ltframe->actual_string = NULL;
    
    if (fread(&ltframe->encoding, 1, 1, fp) != 1) {
        return 0;
    }
    nread = 1;
    spDebug(80, "readID3LangTextFrame", "encoding = %d\n", ltframe->encoding);

    if (lang_flag) {
        if (fread(ltframe->lang, 1, 3, fp) != 3) {
            return 0;
        }
        nread += 3;
        spDebug(80, "readID3LangTextFrame", "lang = %c%c%c\n", ltframe->lang[0], ltframe->lang[1], ltframe->lang[2]);
    }

    len = ltframe->header.size - (long)nread;
    ltframe->string = xspAlloc(len + 2, unsigned char);

    if (header->version == 4) {
        need_len = 0;
        read_len = len;
    } else {
        need_len = len;
        read_len = 0;
    }

    uc = 0;
    if ((len = spReadID3String(ltframe->encoding, ltframe->string, need_len, read_len, &ltframe->bom_flag, &ltframe->swap_flag,
                               spIsID3FrameUnsynchronized((spID3Frame *)ltframe), &last_value, &uc, fp)) <= 0) {
        spDebug(10, "readID3LangTextFrame", "Can't read string.\n");
        return 0;
    }
    ltframe->string[len - uc] = 0;
    ltframe->string[len - uc + 1] = 0;
    header->unsync_count += uc;
    nread += len;
    spDebug(80, "readID3LangTextFrame", "string = `%s'\n", ltframe->string);

    if (strneq(ltframe->header.type, "COM ", 4)
        || strneq(ltframe->header.type, "COMM", 4)
        || strneq(ltframe->header.type, "ULT ", 4)
        || strneq(ltframe->header.type, "USLT", 4)
        || strneq(ltframe->header.type, "TXX ", 4)
        || strneq(ltframe->header.type, "TXXX", 4)
        || strneq(ltframe->header.type, "WXX ", 4)
        || strneq(ltframe->header.type, "WXXX", 4)
        ) {
        int i;

        for (i = (int)strlen((char *)ltframe->string) + 1; i < len - uc; i++) {
            if (ltframe->string[i] != 0) {
                ltframe->actual_string = &ltframe->string[i];
                break;
            }
        }
    }
    
    spDebug(80, "readID3LangTextFrame", "done: nread = %ld / %ld\n", (long)nread, ltframe->header.size);
    
    return nread;
}

spID3FrameSize spReadID3DescTextFrame(spChunk *parent, long depth, void *frame, int swap, FILE *fp)
{
    return readID3LangTextFrame(parent, depth, frame, swap, fp, SP_FALSE);
}

spID3FrameSize spReadID3LangTextFrame(spChunk *parent, long depth, void *frame, int swap, FILE *fp)
{
    return readID3LangTextFrame(parent, depth, frame, swap, fp, SP_TRUE);
}

static spID3FrameSize writeID3LangTextFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp, spBool lang_flag)
{
    long len;
    long write_len;
    long uc;
    unsigned char last_value = 0;
    spID3FrameSize nwrite;
    spID3Header *header;
    spID3LangTextFrame *ltframe = (spID3LangTextFrame *)frame;

    spDebug(80, "writeID3LangTextFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)frame);
    
    spDebug(80, "writeID3LangTextFrame", "encoding = %d\n", ltframe->encoding);
    if (fwrite(&ltframe->encoding, 1, 1, fp) != 1) {
        return 0;
    }
    nwrite = 1;

    if (lang_flag) {
        spDebug(80, "writeID3LangTextFrame", "lang = %c%c%c\n", ltframe->lang[0], ltframe->lang[1], ltframe->lang[2]);
        if (fwrite(ltframe->lang, 1, 3, fp) != 3) {
            return 0;
        }
        nwrite += 3;
    }

    write_len = ltframe->header.size - (long)nwrite;
    if (ltframe->bom_flag) {
        write_len -= 2;
    }

    spDebug(80, "writeID3LangTextFrame", "write_len = %ld, string = `%s'\n", write_len, ltframe->string);
    uc = 0;
    if ((len = spWriteID3String(ltframe->encoding, ltframe->string, write_len, ltframe->bom_flag, ltframe->swap_flag,
                               spIsID3FrameUnsynchronized((spID3Frame *)ltframe), &last_value, &uc, fp)) <= 0) {
        spDebug(10, "writeID3LangTextFrame", "Can't write string.\n");
        return 0;
    }
    header->unsync_count += uc;
    spUpdateID3FrameUnsynchronizedFlag((spID3Frame *)ltframe);
    nwrite += len;

    spDebug(80, "writeID3LangTextFrame", "done: nwrite = %ld / %ld, unsync_count = %ld\n",
            (long)nwrite, ltframe->header.size, header->unsync_count);
    
    return nwrite;
}

spID3FrameSize spWriteID3DescTextFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    return writeID3LangTextFrame(frame, depth, rewrite_flag, swap, fp, SP_FALSE);
}

spID3FrameSize spWriteID3LangTextFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    return writeID3LangTextFrame(frame, depth, rewrite_flag, swap, fp, SP_TRUE);
}

spBool spFreeID3TextFrame(void *frame)
{
    spID3TextFrame *tframe = (spID3TextFrame *)frame;

    if (tframe->string != NULL) {
        xspFree(tframe->string);
        tframe->string = NULL;
    }
    
    return SP_TRUE;
}

static long copyID3String(unsigned char **dest_pstring, long *dest_len, unsigned char **src_pstring, long src_len)
{
    long ncopy = 0;
    unsigned char *dest_string = NULL;
    
    if (*src_pstring != NULL) {
        *dest_len = src_len;
        dest_string = xspAlloc(src_len + 2, unsigned char);
        memcpy(dest_string, *src_pstring, src_len);
        dest_string[src_len] = NUL;
        dest_string[src_len + 1] = NUL;        /* just in case */
        ncopy += src_len;
    } else {
        *dest_len = 0;
    }

    *dest_pstring = dest_string;

    ncopy += sizeof(long) + sizeof(unsigned char *);

    return ncopy;
}

spID3FrameSize spCopyID3TextFrame(spChunk *dest_frame, spChunk *src_frame, long depth)
{
    long dest_len = 0;
    spID3FrameSize ncopy;
    spID3TextFrame *dest_tframe = (spID3TextFrame *)dest_frame;
    spID3TextFrame *src_tframe = (spID3TextFrame *)src_frame;

    dest_tframe->encoding = src_tframe->encoding;
    dest_tframe->bom_flag = src_tframe->bom_flag;
    dest_tframe->swap_flag = src_tframe->swap_flag;
    ncopy = sizeof(dest_tframe->encoding) + sizeof(dest_tframe->bom_flag) + sizeof(dest_tframe->swap_flag);

    ncopy += copyID3String(&dest_tframe->string, &dest_len, &src_tframe->string, src_tframe->header.size - 1);
    
    return ncopy;
}

static spID3FrameSize copyID3LangTextFrame(spChunk *dest_frame, spChunk *src_frame, long depth, spBool lang_flag)
{
    long len;
    spID3FrameSize ncopy;
    spID3LangTextFrame *dest_ltframe = (spID3LangTextFrame *)dest_frame;
    spID3LangTextFrame *src_ltframe = (spID3LangTextFrame *)src_frame;

    spDebug(80, "copyID3LangTextFrame", "in\n");

    dest_ltframe->string = NULL;
    dest_ltframe->actual_string = NULL;

    dest_ltframe->encoding = src_ltframe->encoding;
    len = src_ltframe->header.size - 1;
    ncopy = sizeof(dest_ltframe->encoding);

    dest_ltframe->bom_flag = src_ltframe->bom_flag;
    dest_ltframe->swap_flag = src_ltframe->swap_flag;
    ncopy += sizeof(dest_ltframe->bom_flag) + sizeof(dest_ltframe->swap_flag);
    
    if (lang_flag) {
        memcpy(dest_ltframe->lang, src_ltframe->lang, 3);
        ncopy += 3;
        len -= 3;
    }

    dest_ltframe->string = xspAlloc(len + 2, unsigned char);
    memcpy(dest_ltframe->string, src_ltframe->string, len);
    dest_ltframe->string[len] = 0;
    dest_ltframe->string[len + 1] = 0;
    ncopy += len;

    if (strneq(src_ltframe->header.type, "COM ", 4)
        || strneq(src_ltframe->header.type, "COMM", 4)
        || strneq(src_ltframe->header.type, "ULT ", 4)
        || strneq(src_ltframe->header.type, "USLT", 4)
        || strneq(src_ltframe->header.type, "TXX ", 4)
        || strneq(src_ltframe->header.type, "TXXX", 4)
        || strneq(src_ltframe->header.type, "WXX ", 4)
        || strneq(src_ltframe->header.type, "WXXX", 4)
        ) {
        int i;

        for (i = (int)strlen((char *)dest_ltframe->string) + 1; i < len; i++) {
            if (dest_ltframe->string[i] != 0) {
                dest_ltframe->actual_string = &dest_ltframe->string[i];
                break;
            }
        }
    }
    
    return ncopy;
}

spID3FrameSize spCopyID3DescTextFrame(spChunk *dest_frame, spChunk *src_frame, long depth)
{
    return copyID3LangTextFrame(dest_frame, src_frame, depth, SP_FALSE);
}

spID3FrameSize spCopyID3LangTextFrame(spChunk *dest_frame, spChunk *src_frame, long depth)
{
    return copyID3LangTextFrame(dest_frame, src_frame, depth, SP_TRUE);
}

long spReadUnknownLengthString(unsigned char **pstring, long alloc_incr,
                               long remain_size, long read_limit,
                               int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    unsigned char c;
    long len;
    long nread;
    long nchar;
    long alloc_len;
    unsigned char *string;

    alloc_incr = MAX(alloc_incr, 16);
    alloc_len = alloc_incr;
    
    *pstring = xspAlloc(alloc_len, unsigned char);
    string = *pstring;

    for (nchar = 0, nread = 0;;) {
        if ((len = spReadID3UnsynchronizedBuffer(&c, 1, 0, 1, 1, SP_FALSE, unsync_flag, last_value, unsync_count, fp)) <= 0) {
            return 0;
        }
        nread += len;

        if (c == NUL) {
            break;
        }

        if (remain_size > 0 && ++nchar >= remain_size) {
            break;
        }
        if (read_limit > 0 && nread >= read_limit) {
            break;
        }

        if (nchar + 2 >= alloc_len) {
            alloc_len += alloc_incr;
            *pstring = xrealloc(*pstring, alloc_len, unsigned char);
            string = *pstring;
        }
        string[nchar - 1] = c;
    }
    
    string[nchar] = NUL;
    string[nchar + 1] = NUL;        /* just in case */

    return nread;
}

spID3FrameSize spReadID3MimeDataFrame(spChunk *parent, long depth, void *frame, int swap, FILE *fp)
{
    long len;
    long uc;
    long remain_size, read_limit;
    unsigned char last_value = 0;
    spBool unsynchronized;
    long nread;
    spID3Header *header;
    spID3MimeDataFrame *mdframe = (spID3MimeDataFrame *)frame;

    spDebug(80, "spReadID3MimeDataFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)parent);
    unsynchronized = spIsID3FrameUnsynchronized((spID3Frame *)mdframe);
    
    memset(mdframe->format, 0, 3);
    mdframe->mime_type_len = 0;
    mdframe->mime_type = NULL;
    mdframe->filename_len = 0;
    mdframe->filename = NULL;
    mdframe->data_type = 0;
    mdframe->description_len = 0;
    mdframe->description = NULL;
    mdframe->data_size = 0;
    mdframe->data = NULL;
    
    if (fread(&mdframe->encoding, 1, 1, fp) != 1) {
        return 0;
    }
    nread = 1;
    spDebug(80, "spReadID3MimeDataFrame", "encoding = %d\n", mdframe->encoding);

    if (strneq(mdframe->header.type, "PIC ", 4)) {
        if (fread(mdframe->format, 1, 3, fp) != 3) {
            return 0;
        }
        nread += 3;
        spDebug(80, "spReadID3MimeDataFrame", "format = %c%c%c\n", mdframe->format[0], mdframe->format[1], mdframe->format[2]);
    } else {
        if (header->version == 4) {
            remain_size = 0;
            read_limit = mdframe->header.size - nread;
        } else {
            remain_size = mdframe->header.size - nread;
            read_limit = 0;
        }

        uc = 0;
        if ((len = spReadUnknownLengthString(&mdframe->mime_type, 16, remain_size, read_limit,
                                             unsynchronized, &last_value, &uc, fp)) <= 0) {
            spDebug(80, "spReadID3MimeDataFrame", "spReadUnknownLengthString failed\n");
            return 0;
        }
        header->unsync_count += uc;
        mdframe->mime_type_len = len - uc;
        nread += len;
        spDebug(80, "spReadID3MimeDataFrame", "len = %d, mime_type = %s\n", len, mdframe->mime_type);
    }

    if (strneq(mdframe->header.type, "GEO", 3)) {
        if (header->version == 4) {
            remain_size = 0;
            read_limit = mdframe->header.size - nread;
        } else {
            remain_size = mdframe->header.size - nread;
            read_limit = 0;
        }

        uc = 0;
        if ((len = spReadUnknownLengthString(&mdframe->filename, 16, remain_size, read_limit,
                                             unsynchronized, &last_value, &uc, fp)) <= 0) {
            spDebug(80, "spReadID3MimeDataFrame", "spReadUnknownLengthString failed\n");
            return 0;
        }
        header->unsync_count += uc;
        mdframe->filename_len = len - uc;
        nread += len;
        spDebug(80, "spReadID3MimeDataFrame", "len = %d, filename = %s\n", len, mdframe->filename);
    } else {
        if (fread(&mdframe->data_type, 1, 1, fp) != 1) {
            return 0;
        }
        nread += 1;
        spDebug(80, "spReadID3MimeDataFrame", "data_type = %d\n", mdframe->data_type);
    }

    if (header->version == 4) {
        remain_size = 0;
        read_limit = mdframe->header.size - nread;
    } else {
        remain_size = mdframe->header.size - nread;
        read_limit = 0;
    }

    uc = 0;
    if ((len = spReadUnknownLengthString(&mdframe->description, 16, remain_size, read_limit,
                                         unsynchronized, &last_value, &uc, fp)) <= 0) {
        spDebug(80, "spReadID3MimeDataFrame", "spReadUnknownLengthString failed\n");
        return 0;
    }
    header->unsync_count += uc;
    mdframe->description_len = len - uc;
    nread += len;
    spDebug(80, "spReadID3MimeDataFrame", "len = %d, description = %s\n", len, mdframe->description);

    mdframe->data_size = mdframe->header.size - nread;
    mdframe->data = xspAlloc(mdframe->data_size, unsigned char);

    if (unsynchronized == SP_TRUE) {
        spDebug(10, "spReadID3MimeDataFrame", "**** unsynchrozisation ****\n");

        if (header->version == 4) {
            remain_size = 0;
            read_limit = mdframe->data_size;
        } else {
            remain_size = mdframe->data_size;
            read_limit = 0;
        }
    
        uc = 0;
        if ((len = spReadID3UnsynchronizedBuffer(mdframe->data, remain_size, read_limit, 1, 1, SP_FALSE,
                                                 SP_TRUE, &last_value, &uc, fp)) <= 0) {
            spDebug(10, "spReadID3MimeDataFrame", "spReadID3UnsynchronizedBuffer failed\n");
            return 0;
        }
        if (header->version == 4) {
            mdframe->data_size -= uc;
        }
        header->unsync_count += uc;
        nread += len;
    } else {
        spDebug(10, "spReadID3MimeDataFrame", "---- no unsynchrozisation ----\n");
        if (fread(mdframe->data, 1, mdframe->data_size, fp) != mdframe->data_size) {
            return 0;
        }
        nread += mdframe->data_size;
    }
    
    spDebug(80, "spReadID3MimeDataFrame", "data_size = %ld\n", mdframe->data_size);

    spDebug(80, "spReadID3MimeDataFrame", "done: nread = %ld / %ld\n", (long)nread, mdframe->header.size);
    
    return nread;
}

long spWriteUnknownLengthString(unsigned char *string, long len, int unsync_flag, unsigned char *last_value, long *unsync_count, FILE *fp)
{
    return spWriteID3UnsynchronizedBuffer(string, len, 1, 1, SP_FALSE, unsync_flag, last_value, unsync_count, fp);
}

spID3FrameSize spWriteID3MimeDataFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    long len;
    long uc;
    unsigned char last_value = 0;
    spID3FrameSize nwrite;
    spID3Header *header;
    spID3MimeDataFrame *mdframe = (spID3MimeDataFrame *)frame;

    spDebug(80, "spWriteID3MimeDataFrame", "in\n");

    header = spGetID3HeaderFromFrame((spID3Frame *)frame);
    
    spDebug(80, "spWriteID3MimeDataFrame", "encoding = %d\n", mdframe->encoding);
    if (fwrite(&mdframe->encoding, 1, 1, fp) != 1) {
        return 0;
    }
    nwrite = 1;

    if (strneq(mdframe->header.type, "PIC ", 4)) {
        spDebug(80, "spWriteID3MimeDataFrame", "format = %c%c%c\n", mdframe->format[0], mdframe->format[1], mdframe->format[2]);
        if (fwrite(mdframe->format, 1, 3, fp) != 3) {
            return 0;
        }
        nwrite += 3;
    } else {
        spDebug(80, "spWriteID3MimeDataFrame", "len = %d, mime_type = %s\n", mdframe->mime_type_len, mdframe->mime_type);
        uc = 0;
        if ((len = spWriteUnknownLengthString(mdframe->mime_type, mdframe->mime_type_len, 
                                              SP_TRUE, &last_value, &uc, fp)) <= 0) {
            spDebug(80, "spWriteID3MimeDataFrame", "spWriteUnknownLengthString failed\n");
            return 0;
        }
        header->unsync_count += uc;
        nwrite += len;
    }

    if (strneq(mdframe->header.type, "GEO", 3)) {
        spDebug(80, "spWriteID3MimeDataFrame", "len = %d, filename = %s\n", mdframe->filename_len, mdframe->filename);
        uc = 0;
        if ((len = spWriteUnknownLengthString(mdframe->filename, mdframe->filename_len,
                                              SP_TRUE, &last_value, &uc, fp)) <= 0) {
            spDebug(80, "spWriteID3MimeDataFrame", "spWriteUnknownLengthString failed\n");
            return 0;
        }
        header->unsync_count += uc;
        nwrite += len;
    } else {
        spDebug(80, "spWriteID3MimeDataFrame", "data_type = %d\n", mdframe->data_type);
        if (fwrite(&mdframe->data_type, 1, 1, fp) != 1) {
            return 0;
        }
        nwrite += 1;
    }

    spDebug(80, "spWriteID3MimeDataFrame", "len = %d, description = %s\n", mdframe->description_len, mdframe->description);
    uc = 0;
    if ((len = spWriteUnknownLengthString(mdframe->description, mdframe->description_len,
                                          SP_TRUE, &last_value, &uc, fp)) <= 0) {
        spDebug(80, "spWriteID3MimeDataFrame", "spWriteUnknownLengthString failed\n");
        return 0;
    }
    header->unsync_count += uc;
    nwrite += len;

    spDebug(80, "spWriteID3MimeDataFrame", "data_size = %ld\n", mdframe->data_size);
    uc = 0;
    if ((len = spWriteID3UnsynchronizedBuffer(mdframe->data, mdframe->data_size, 1, 1, SP_FALSE,
                                              SP_TRUE, &last_value, &uc, fp)) <= 0) {
        spDebug(10, "spWriteID3MimeDataFrame", "spWriteID3UnsynchronizedBuffer failed\n");
        return 0;
    }
    header->unsync_count += uc;
    nwrite += len;
    
    spUpdateID3FrameUnsynchronizedFlag((spID3Frame *)mdframe);
    
    spDebug(80, "spWriteID3MimeDataFrame", "done: nwrite = %ld / %ld\n", (long)nwrite, mdframe->header.size);
    
    return nwrite;
}

spID3FrameSize spCopyID3MimeDataFrame(spChunk *dest_frame, spChunk *src_frame, long depth)
{
    long ncopy;
    spID3MimeDataFrame *dest_mdframe = (spID3MimeDataFrame *)dest_frame;
    spID3MimeDataFrame *src_mdframe = (spID3MimeDataFrame *)src_frame;

    spDebug(80, "spCopyID3MimeDataFrame", "in\n");

    dest_mdframe->encoding = src_mdframe->encoding;
    ncopy = 1;

    memcpy(dest_mdframe->format, src_mdframe->format, 3);
    ncopy += 3;
    
    ncopy += copyID3String(&dest_mdframe->mime_type, &dest_mdframe->mime_type_len,
                           &src_mdframe->mime_type, src_mdframe->mime_type_len);

    ncopy += copyID3String(&dest_mdframe->filename, &dest_mdframe->filename_len,
                           &src_mdframe->filename, src_mdframe->filename_len);

    dest_mdframe->data_type = src_mdframe->data_type;
    ncopy += 1;

    ncopy += copyID3String(&dest_mdframe->description, &dest_mdframe->description_len,
                           &src_mdframe->description, src_mdframe->description_len);

    dest_mdframe->data_size = src_mdframe->data_size;
    dest_mdframe->data = xspAlloc(dest_mdframe->data_size, unsigned char);
    memcpy(dest_mdframe->data, src_mdframe->data, dest_mdframe->data_size);
    ncopy += dest_mdframe->data_size;
    
    return ncopy;
}

spBool spFreeID3MimeDataFrame(void *frame)
{
    spID3MimeDataFrame *mdframe = (spID3MimeDataFrame *)frame;

    if (mdframe->mime_type != NULL) {
        xspFree(mdframe->mime_type);
        mdframe->mime_type = NULL;
    }
    
    if (mdframe->filename != NULL) {
        xspFree(mdframe->filename);
        mdframe->filename = NULL;
    }
    
    if (mdframe->description != NULL) {
        xspFree(mdframe->description);
        mdframe->description = NULL;
    }
    
    if (mdframe->data != NULL) {
        xspFree(mdframe->data);
        mdframe->data = NULL;
    }
    
    return SP_TRUE;
}

static spChunkInfoTable sp_id3_frame_info_list[] = {
    {"", "ID3", SP_FALSE, SP_CHUNK_INFO_OPTION_MARGIN_CHUNK_MASK | SP_CHUNK_INFO_OPTION_ACCEPT_ZERO_TYPE_MASK, 0, 0, 0, NULL, NULL, NULL, NULL, SP_TRUE, NULL, NULL},
    {"TXX ", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_DESC_TEXT_MASK, 0, sizeof(spID3DescTextFrame), 0, spReadID3DescTextFrame, spWriteID3DescTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3DescTextFrame},
    {"TXXX", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_DESC_TEXT_MASK, 0, sizeof(spID3DescTextFrame), 0, spReadID3DescTextFrame, spWriteID3DescTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3DescTextFrame},
    {"WXX ", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_DESC_TEXT_MASK, 0, sizeof(spID3DescTextFrame), 0, spReadID3DescTextFrame, spWriteID3DescTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3DescTextFrame},
    {"WXXX", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_DESC_TEXT_MASK, 0, sizeof(spID3DescTextFrame), 0, spReadID3DescTextFrame, spWriteID3DescTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3DescTextFrame},
    {"T", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_TEXT_MASK, 0, sizeof(spID3TextFrame), 0, spReadID3TextFrame, spWriteID3TextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3TextFrame},
    {"W", "ID3", SP_FALSE, 0, 0, sizeof(spID3DataFrame), 0, spReadID3DataFrame, spWriteID3DataFrame, NULL, spFreeID3DataFrame, SP_TRUE, NULL, spCopyID3DataFrame},
    {"UFI", "ID3", SP_FALSE, 0, 0, sizeof(spID3DataFrame), 0, spReadID3DataFrame, spWriteID3DataFrame, NULL, spFreeID3DataFrame, SP_TRUE, NULL, spCopyID3DataFrame},
    {"COM ", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_LANG_TEXT_MASK, 0, sizeof(spID3LangTextFrame), 0, spReadID3LangTextFrame, spWriteID3LangTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3LangTextFrame},
    {"COMM", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_LANG_TEXT_MASK, 0, sizeof(spID3LangTextFrame), 0, spReadID3LangTextFrame, spWriteID3LangTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3LangTextFrame},
    {"USER", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_LANG_TEXT_MASK, 0, sizeof(spID3LangTextFrame), 0, spReadID3LangTextFrame, spWriteID3LangTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3LangTextFrame},
    {"ULT ", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_LANG_TEXT_MASK, 0, sizeof(spID3LangTextFrame), 0, spReadID3LangTextFrame, spWriteID3LangTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3LangTextFrame},
    {"USLT", "ID3", SP_FALSE, SP_ID3_CHUNK_INFO_OPTION_LANG_TEXT_MASK, 0, sizeof(spID3LangTextFrame), 0, spReadID3LangTextFrame, spWriteID3LangTextFrame, NULL, spFreeID3TextFrame, SP_TRUE, NULL, spCopyID3LangTextFrame},
    {"PIC ", "ID3", SP_FALSE, 0, 0, sizeof(spID3MimeDataFrame), 0, spReadID3MimeDataFrame, spWriteID3MimeDataFrame, NULL, spFreeID3MimeDataFrame, SP_TRUE, NULL, spCopyID3MimeDataFrame},
    {"APIC", "ID3", SP_FALSE, 0, 0, sizeof(spID3MimeDataFrame), 0, spReadID3MimeDataFrame, spWriteID3MimeDataFrame, NULL, spFreeID3MimeDataFrame, SP_TRUE, NULL, spCopyID3MimeDataFrame},
    {"GEO", "ID3", SP_FALSE, 0, 0, sizeof(spID3MimeDataFrame), 0, spReadID3MimeDataFrame, spWriteID3MimeDataFrame, NULL, spFreeID3MimeDataFrame, SP_TRUE, NULL, spCopyID3MimeDataFrame},
    {"PRIV", "ID3", SP_FALSE, 0, 0, sizeof(spID3DataFrame), 0, spReadID3DataFrame, spWriteID3DataFrame, NULL, spFreeID3DataFrame, SP_TRUE, NULL, spCopyID3DataFrame},
};

spChunkFileSpec sp_id3_file_spec = {
    sizeof(struct _spID3FrameHeader),
    sizeof(struct _spID3Frame),
    sp_id3_frame_info_list,
    0,

    NULL,
    NULL,
    NULL,
    (spChunkGetSizeFunc)spGetID3FrameHeaderSize,
    (spChunkGetSizeFunc)spGetID3FrameContentSizeForChunkFileSpec,
    (spChunkSetSizeFunc)spSetID3FrameContentSizeForChunkFileSpec,
    NULL,
    NULL,
    (spChunkReadHeaderFunc)spReadID3FrameHeader,
    NULL,
    (spChunkSkipFunc)spSkipID3Frame,
        
    (spChunkWriteHeaderFunc)spWriteID3FrameHeader,
    NULL,
    (spChunkWriteContentRestFunc)spWriteID3ContentRest,
        
    NULL,
};

spBool spInitID3Header(spID3Header *header, unsigned char version, unsigned char revision)
{
    if (header == NULL || !(version == 2 || version == 3 || version == 4)) {
        return SP_FALSE;
    }
    
    memset(header, 0, sizeof(spID3Header));
    header->version = version;
    header->revision = revision;
    header->header.type[0] = 'I';
    header->header.type[1] = 'D';
    header->header.type[2] = '3';
    header->header.type[3] = NUL;

    return SP_TRUE;
}

spID3Frame *spAppendID3Frame(spID3Header *header, const char *type, spBool find_existing_frame)
{
    spID3Frame *frame = NULL;

    if (find_existing_frame == SP_FALSE || (frame = spFindID3FrameFromHeader(header, type)) == NULL) {
        spDebug(80, "spAppendID3Frame", "call spCreateID3Frame\n");
        if ((frame = spCreateID3Frame(header, type)) == NULL) {
            return NULL;
        }
    }

    return (spID3Frame *)frame;
}

void spPropagateID3FrameContentSize(spID3Frame *parent, spChunkSize size_diff)
{
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }
    
    spPropagateChunkContentSize(&sp_id3_file_spec, (spChunk *)parent, size_diff);

    return;
}

spBool spSetID3FrameContentSize(spID3Frame *frame, spID3FrameSize size, spBool propagate_size)
{
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }
    spDebug(80, "spSetID3FrameContentSize", "size = %lu, propagate_size = %d\n", (unsigned long)size, propagate_size);
    return spSetChunkContentSize(&sp_id3_file_spec, (spChunk *)frame, size, propagate_size);
}

spID3Frame *spCreateID3FrameWithCB(spID3Header *header, const char *type, spID3FrameUpdateFunc update_func, void *data)
{
    const char *vtype;
    spID3Frame *frame;
    
    if (header == NULL || strnone(type)) return NULL;
    
    spDebug(80, "spCreateID3FrameWithCB", "type = %s\n", type);
    
    if ((vtype = spGetID3HeaderCurrentVersionTypeId(header, type)) != NULL) {
        spDebug(80, "spCreateID3FrameWithCB", "vtype = %s\n", vtype);
        type = vtype;
    }
    
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }

    if ((frame = (spID3Frame *)spCreateChunk(&sp_id3_file_spec, (spChunk *)header, NULL, type,
                                             SP_TRUE, SP_TRUE, (spChunkUpdateFunc)update_func, data)) == NULL) {
        return NULL;
    }

    return frame;
}

spID3Frame *spCreateID3Frame(spID3Header *header, const char *type)
{
    return spCreateID3FrameWithCB(header, type, NULL, NULL);
}

spID3FrameSize spReadID3ChildFrame(spChunk *parent, long max_count, spID3FrameSize remain_size, long depth, int swap, FILE *fp)
{
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }
    return spReadChildChunk(&sp_id3_file_spec, parent, max_count, remain_size, depth, swap, fp);
}

spID3FrameSize spReadID3Header(spID3Header *header, FILE *fp)
{
    spID3FrameSize nread;
    spID3FrameSize total_nread;

    memset(header, 0, sizeof(spID3Header));

    if ((nread = fread(header->header.type, 1, 3, fp)) != 3
        || !strneq(header->header.type, "ID3", 3)) {
        if (nread > 0) {
            spSeekFile(fp, -nread, SEEK_CUR);
        }
        spDebug(10, "spReadID3Header", "Can't read first 'ID3': %ld\n", (long)nread);
        return 0;
    }
    total_nread = nread;

    if ((nread = fread(&header->version, 1, 1, fp)) != 1) {
        spDebug(10, "spReadID3Header", "Can't read version: %ld\n", (long)nread);
        return 0;
    }
    spDebug(80, "spReadID3Header", "header->version = %d\n", header->version);
    total_nread += 1;

    if (header->version == 2 || header->version == 3 || header->version == 4) {
        spDebug(80, "spReadID3Header", "supported version: %d\n", header->version);
    } else {
        spDebug(10, "spReadID3Header", "unsupported version: %d\n", header->version);
        return 0;
    }
    
    if ((nread = fread(&header->revision, 1, 1, fp)) != 1) {
        spDebug(10, "spReadID3Header", "Can't read revision: %ld\n", (long)nread);
        return 0;
    }
    spDebug(80, "spReadID3Header", "header->revision = %d\n", header->revision);
    total_nread += 1;
    
    if ((nread = fread(&header->header.flag1, 1, 1, fp)) != 1) {
        spDebug(10, "spReadID3Header", "Can't read flag: %ld\n", (long)nread);
        return 0;
    }
    spDebug(80, "spReadID3Header", "header->header.flag1 = %x\n", header->header.flag1);
    total_nread += 1;
    
    if ((nread = spReadID3SyncSafeLong32(&header->header.size, fp)) != 4) {
        spDebug(10, "spReadID3Header", "Can't read size: %ld\n", (long)nread);
        return 0;
    }
    spDebug(80, "spReadID3Header", "header->header.size = %ld\n", (long)header->header.size);
    total_nread += nread;
    
    if ((nread = spReadID3ChildFrame((spChunk *)header, 0, header->header.size, 0, SP_AIFF_NEED_SWAP, fp)) <= 0) {
        spDebug(10, "spReadID3Header", "spReadID3ChildFrame failed: %ld\n", (long)nread);
        return 0;
    }
    spDebug(80, "spReadID3Header", "spReadID3ChildFrame result: nread = %ld\n", (long)nread);
    total_nread += nread;

    spDebug(80, "spReadID3Header", "done: total_nread = %ld\n", (long)total_nread);
    
    return total_nread;
}

spChunkSize spWriteID3ChildFrame(spChunk *frame, long depth, spBool rewrite_flag, int swap, FILE *fp)
{
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }
    return spWriteChildChunk(&sp_id3_file_spec, frame, depth, SP_TRUE, rewrite_flag, swap, NULL, fp);
}

spID3FrameSize spWriteID3Header(spID3Header *header, FILE *fp)
{
    spID3FrameSize nwrite = 0;
    spID3FrameSize total_nwrite = 0;

    if (!strneq(header->header.type, "ID3", 3)
        || !(header->version == 2 || header->version == 3 || header->version == 4)) {
        spDebug(10, "spWriteID3Header", "not supported: version = %d\n", header->version);
        return 0;
    }

    if ((nwrite = fwrite(header->header.type, 1, 3, fp)) == 3) {
        total_nwrite = nwrite;
        if ((nwrite = fwrite(&header->version, 1, 1, fp)) == 1) {
            total_nwrite += nwrite;
            spDebug(80, "spWriteID3Header", "write revision: current total_nwrite = %ld, revision = %d\n",
                    (long)total_nwrite, header->revision);
            if ((nwrite = fwrite(&header->revision, 1, 1, fp)) == 1) {
                total_nwrite += nwrite;
                spDebug(80, "spWriteID3Header", "write flag1: current total_nwrite = %ld, flag1 = %x\n",
                        (long)total_nwrite, header->header.flag1);
                if ((nwrite = fwrite(&header->header.flag1, 1, 1, fp)) == 1) {
                    total_nwrite += nwrite;
                    spDebug(80, "spWriteID3Header", "call spWriteID3SyncSafeLong32: current total_nwrite = %ld, header->header.size = %ld\n",
                            (long)total_nwrite, (long)header->header.size);
                    if ((nwrite = spWriteID3SyncSafeLong32(header->header.size, fp)) == 4) {
                        total_nwrite += nwrite;
                    }
                }
            }
        }
    }

    spDebug(80, "spWriteID3Header", "write header: total_nwrite = %ld\n", (long)total_nwrite);
    
    if (total_nwrite != 10) {
        spDebug(10, "spWriteID3Header", "Can't write ID3 header: %ld\n", (long)nwrite);
        return 0;
    }

    if ((nwrite = spWriteID3ChildFrame((spChunk *)header, 0, SP_FALSE, SP_AIFF_NEED_SWAP, fp)) > 0) {
        total_nwrite += nwrite;
        spDebug(80, "spWriteID3Header", "done: total_nwrite = %ld\n", (long)total_nwrite);
        return total_nwrite;
    } else {
        spDebug(10, "spWriteID3Header", "spWriteID3ChildFrame failed: %ld\n", (long)nwrite);
        return 0;
    }
}

spBool spFreeID3Header(spID3Header *header)
{
    spDebug(80, "spFreeID3Header", "in\n");
    
    if (header != NULL) {
        return spFreeChunk(&sp_id3_file_spec, (spChunk *)header);
    } else {
        spDebug(80, "spFreeID3Header", "header == NULL\n");
        return SP_FALSE;
    }
}

spID3FrameSize spCopyID3ChildFrame(spChunk *dest_parent, spChunk *src_parent, long depth)
{
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }
    return spCopyChildChunk(&sp_id3_file_spec, dest_parent, src_parent, depth);
}

spID3FrameSize spCopyID3Header(spID3Header *dest_header, spID3Header *src_header)
{
    spID3FrameSize ncopy;
    spID3FrameSize total_ncopy;

    memcpy(dest_header, src_header, sizeof(spID3Header));
    dest_header->child = NULL;

    if ((ncopy = spCopyID3ChildFrame((spChunk *)dest_header, (spChunk *)src_header, 0)) <= 0) {
        return ncopy;
    }
    total_ncopy = ncopy;

    spDebug(80, "spCopyID3Header", "done: total_ncopy = %ld\n", (long)total_ncopy);
    
    return total_ncopy;
}

spChunkInfoTable *spFindID3FrameInfoTable(spID3FrameHeader *parent_header, spID3FrameHeader *header)
{
    if (sp_id3_file_spec.num_list <= 0) {
        sp_id3_file_spec.num_list = spArraySize(sp_id3_frame_info_list);
    }

    return spFindChunkInfoTable(sp_id3_frame_info_list, sp_id3_file_spec.num_list, (spChunkHeader *)parent_header, (spChunkHeader *)header);
}

spID3Frame *spFindID3FrameFromHeader(spID3Header *header, const char *type)
{
    const char *vtype;
    
    spDebug(80, "spFindID3FrameFromHeader", "type = %s\n", type);
    
    if ((vtype = spGetID3HeaderCurrentVersionTypeId(header, type)) == NULL) {
        spDebug(80, "spFindID3FrameFromHeader", "cannot find current version type: type = %s\n", type);
        return NULL;
    }
    spDebug(80, "spFindID3FrameFromHeader", "vtype = %s\n", vtype);
    
    return (spID3Frame *)spFindChildChunk((spChunk *)header, vtype, "ID3");
}

spID3Frame *spFindID3Frame(spID3Frame *frame, const char *type)
{
    const char *vtype;
    spID3Header *header;

    if ((header = spGetID3HeaderFromFrame(frame)) == NULL) {
        return NULL;
    }
    
    if ((vtype = spGetID3HeaderCurrentVersionTypeId(header, type)) == NULL) {
        return NULL;
    }
    
    return (spID3Frame *)spFindChunk((spChunk *)frame, type, "ID3");
}

const unsigned char *spGetID3TextFrameString(spID3TextFrame *tframe, unsigned char *encoding)
{
    spChunkInfoTable *info;
    unsigned char *string = NULL;

    if (tframe == NULL || tframe->parent == NULL) return NULL;

    if ((info = spFindID3FrameInfoTable(&tframe->parent->header, &tframe->header)) == NULL) {
        return NULL;
    }
    
    if (info->options & SP_ID3_CHUNK_INFO_OPTION_TEXT_MASK) {
        if (info->options & SP_ID3_CHUNK_INFO_OPTION_DESC_MASK) {
            spID3DescTextFrame *dtframe = (spID3DescTextFrame *)tframe;

            string = dtframe->actual_string;
        } else {
            string = tframe->string;
        }

        if (encoding != NULL) *encoding = tframe->encoding;
    }

    return string;
}

static void convertUTF8StringToLocaleCode(spConverter conv_from_id3, unsigned char *str_in_locale, long size)
{
    if (conv_from_id3 != NULL) {
        char *ostr;
        /* UTF8 --> system codeset */
        if ((ostr = xspConvert(conv_from_id3, (char *)str_in_locale)) != NULL) {
            spDebug(80, "spGetID3TextFrameStringToLocaleCode", "xspConvert result: %s\n", ostr);
            spStrCopy((char *)str_in_locale, size, ostr);
            xspFree(ostr);
        } else {
            spDebug(10, "spGetID3TextFrameStringToLocaleCode", "xspConvert failed\n");
        }
    } else {
        spKanjiCode ocode;
        /* UTF8 --> locale code */
        ocode = spConvertKanjiToLocaleCode(str_in_locale, size, SP_KANJI_CODE_UTF8);
        spDebug(80, "spGetID3TextFrameStringToLocaleCode", "ocode = %d\n", ocode);
    }

    return;
}

/* return SP_FALSE: not genre frame.
 * ostr can be identical to istr. */
static spBool getID3GenreStringIfGenreFrame(spID3Header *header, spID3TextFrame *tframe, unsigned char *istr, unsigned char *ostr, long size, int *ogenre_index)
{
    int genre_index = -1;
    const char *genre_type;

    if (ogenre_index != NULL) *ogenre_index = -1;

    genre_type = spGetID3HeaderCurrentVersionTypeId(header, SP_ID3_FRAME_TYPE_GENRE);
    spDebug(80, "getID3GenreStringIfGenreFrame", "genre_type = %s, tframe->header.type = %s\n",
            genre_type, tframe->header.type);

    /* istr is genre code */
    if (strneq(tframe->header.type, genre_type, strlen(genre_type))) {
        if (istr[0] == '(') {
            char *ptr;
            char buf[SP_MAX_LINE];
            
            spStrCopy(buf, sizeof(buf), (char *)(istr + 1));
            spDebug(80, "getID3GenreStringIfGenreFrame", "buf can be genre number: %s\n", buf);
            
            if ((ptr = spStrChr(buf, ')')) != NULL) {
                const char *genre_str;
                *ptr = NUL;
                genre_index = atoi(buf);
                spDebug(80, "getID3GenreStringIfGenreFrame", "genre_index = %d\n", genre_index);

                if ((genre_str = spGetID3v1GenreString(genre_index, NULL)) != NULL) {
                    /* final ostr: genre string */
                    spStrCopy((char *)ostr, size, genre_str);
                    if (ogenre_index != NULL) *ogenre_index = genre_index;
                } else {
                    genre_index = -1;
                    spDebug(80, "getID3GenreStringIfGenreFrame", "spGetID3v1GenreString failed\n");
                }
            }
        }

        if (genre_index < 0) {
            if (istr != ostr) {
                /* failed: copy original string */
                spStrCopy((char *)ostr, size, (const char *)istr);
            }
        }
        
        return SP_TRUE;
    }

    return SP_FALSE;
}

/* return SP_FALSE: not genre frame.
 * ostr can be identical to istr. */
static spBool getID3GenreIndexStringIfGenreFrame(spID3Header *header, spID3TextFrame *tframe, const unsigned char *istr, unsigned char *ostr, long size, int *ogenre_index)
{
    int genre_index = -1;
    char buf[SP_MAX_LINE];
    const char *genre_type;

    if (ogenre_index != NULL) *ogenre_index = -1;
    
    genre_type = spGetID3HeaderCurrentVersionTypeId(header, SP_ID3_FRAME_TYPE_GENRE);
    spDebug(80, "getID3GenreIndexStringIfGenreFrame", "genre_type = %s, tframe->header.type = %s\n",
            genre_type, tframe->header.type);

    /* genre frame */
    if (strneq(tframe->header.type, genre_type, strlen(genre_type))) {
        genre_index = spFindID3v1GenreIndex((const char *)istr, SP_ID3V1_GENRE_NONEXTENSION);
        spDebug(80, "getID3GenreIndexStringIfGenreFrame", "genre_index = %d\n", genre_index);
        if (genre_index >= 0) {
            sprintf(buf, "(%d)", genre_index);
            spDebug(80, "getID3GenreIndexStringIfGenreFrame", "final genre number string = %s\n", buf);
            spStrCopy((char *)ostr, size, buf);
            if (ogenre_index != NULL) *ogenre_index = genre_index;
        }

        if (genre_index < 0) {
            if (istr != ostr) {
                /* failed: copy original string */
                spStrCopy((char *)ostr, size, (const char *)istr);
            }
        }
        
        return SP_TRUE;
    }

    return SP_FALSE;
}

static spBool getID3TextFrameStringToLocaleCode(spConverter conv_from_id3, spID3TextFrame *tframe, unsigned char *str_in_locale, long size)
{
    unsigned char encoding;
    const unsigned char *string;

    if ((string = spGetID3TextFrameString(tframe, &encoding)) == NULL) {
        return SP_FALSE;
    }
    spDebug(80, "spGetID3TextFrameStringToLocaleCode", "encoding = %d\n", encoding);

    if (encoding == 1/* UTF-16 */ || encoding == 2/* UTF-16BE */) {
        if (spUnicode16ToUTF8((unsigned short *)string, str_in_locale, size) <= 0) {
            return SP_FALSE;
        }
    } else {
        spStrCopy((char *)str_in_locale, size, (char *)string);
    }

    if (encoding != 0) {
        convertUTF8StringToLocaleCode(conv_from_id3, str_in_locale, size);
    } else {
        spID3Header *header;
        int genre_index = -1;

        if ((header = spGetID3HeaderFromFrame((spID3Frame *)tframe)) == NULL) {
            return SP_FALSE;
        }
        
        getID3GenreStringIfGenreFrame(header, tframe, str_in_locale, str_in_locale, size, &genre_index);
        spDebug(80, "spGetID3TextFrameStringToLocaleCode", "genre_index = %d\n", genre_index);
    }

    spDebug(80, "spGetID3TextFrameStringToLocaleCode", "str_in_locale = %s\n", str_in_locale);
    
    return SP_TRUE;
}

spBool spGetID3TextFrameStringToLocaleCode(spID3TextFrame *tframe, unsigned char *str_in_locale, long size)
{
    return getID3TextFrameStringToLocaleCode(NULL, tframe, str_in_locale, size);
}

spBool spGetID3TextFrameTrackNumber(spID3TextFrame *tframe, spLong32 *track, spLong32 *total)
{
    const unsigned char *string;
    char *p;
    char buf[SP_MAX_LINE];
    
    if ((string = spGetID3TextFrameString(tframe, NULL)) == NULL) return SP_FALSE;

    if (total != NULL) *total = 0;

    spStrCopy(buf, sizeof(buf), (char *)string);

    if ((p = spStrChr(buf, '/')) != NULL) {
        if (total != NULL) *total = atol(p + 1);
        *p = NUL;
    }

    if (track != NULL) *track = atol(buf);

    return SP_TRUE;
}

spBool spGetID3TextFrameNumber(spID3TextFrame *tframe, spLong32 *number)
{
    const unsigned char *string;
    char buf[SP_MAX_LINE];
    
    if ((string = spGetID3TextFrameString(tframe, NULL)) == NULL) return SP_FALSE;

    spStrCopy(buf, sizeof(buf), (const char *)string);

    if (number != NULL) *number = atol(buf);
    
    return SP_TRUE;
}

const unsigned char *spFindID3TextFrameString(spID3Header *header, const char *type, unsigned char *encoding)
{
    spID3TextFrame *tframe;
    
    if ((tframe = (spID3TextFrame *)spFindID3FrameFromHeader(header, type)) == NULL) {
        return NULL;
    }

    return spGetID3TextFrameString(tframe, encoding);
}

static spBool findID3TextFrameStringToLocaleCode(spConverter conv_from_id3, spID3Header *header, const char *type, unsigned char *str_in_locale, long size)
{
    spID3TextFrame *tframe;
    
    if ((tframe = (spID3TextFrame *)spFindID3FrameFromHeader(header, type)) == NULL) {
        spDebug(80, "spFindID3TextFrameStringToLocaleCode", "cannot find frame: type = %s\n", type);
        return SP_FALSE;
    }

    return getID3TextFrameStringToLocaleCode(conv_from_id3, tframe, str_in_locale, size);
}

spBool spFindID3TextFrameStringToLocaleCode(spID3Header *header, const char *type, unsigned char *str_in_locale, long size)
{
    return findID3TextFrameStringToLocaleCode(NULL, header, type, str_in_locale, size);
}

spBool spFindID3TextFrameTrackNumber(spID3Header *header, spLong32 *track, spLong32 *total)
{
    spID3TextFrame *tframe;
    
    if ((tframe = (spID3TextFrame *)spFindID3FrameFromHeader(header, SP_ID3_FRAME_TYPE_TRACK)) == NULL) {
        return SP_FALSE;
    }
    
    return spGetID3TextFrameTrackNumber(tframe, track, total);
}

spBool spFindID3TextFrameDiscNumber(spID3Header *header, spLong32 *disc, spLong32 *total)
{
    spID3TextFrame *tframe;
    
    if ((tframe = (spID3TextFrame *)spFindID3FrameFromHeader(header, SP_ID3_FRAME_TYPE_DISC)) == NULL) {
        return SP_FALSE;
    }
    
    return spGetID3TextFrameTrackNumber(tframe, disc, total);
}

spBool spFindID3TextFrameTempo(spID3Header *header, spLong32 *tempo)
{
    spID3TextFrame *tframe;
    
    if ((tframe = (spID3TextFrame *)spFindID3FrameFromHeader(header, SP_ID3_FRAME_TYPE_TEMPO)) == NULL) {
        return SP_FALSE;
    }
    
    return spGetID3TextFrameNumber(tframe, tempo);
}

spBool spSetID3TextFrameString(spID3TextFrame *tframe, unsigned char encoding, const unsigned char *str, long len)
{
    if (tframe == NULL || str == NULL || len <= 0) return SP_FALSE;

    if (tframe->string != NULL) xspFree(tframe->string);
    
    tframe->encoding = encoding;
    
    tframe->string = xspAlloc(len + 2, unsigned char);
    memcpy(tframe->string, str, len);
    tframe->string[len] = 0;
    tframe->string[len + 1] = 0;
    if (encoding == 1/* UTF-16 */ || encoding == 2/* UTF-16BE */) {
        /* NUL character of short */
        len += 2;
    } else {
        /* NUL character of char */
        len += 1;
    }
    if (tframe->bom_flag) {
        len += 2;
    }
    spDebug(80, "spSetID3TextFrameString", "len = %ld, string = %s\n", len, tframe->string);

    /* +1: encoding */
    spSetID3FrameContentSize((spID3Frame *)tframe, len + 1, SP_TRUE);

    return SP_TRUE;
}

static unsigned char *xconvertLocaleCodeToUTF8String(spConverter conv_to_id3, const unsigned char *str_in_locale, long *osize)
{
    long size;
    unsigned char *utf8str = NULL;
    
    if (conv_to_id3 != NULL) {
        /* system codeset --> UTF8 */
        if ((utf8str = (unsigned char *)xspConvert(conv_to_id3, (char *)str_in_locale)) == NULL) {
            spDebug(80, "spSetID3TextFrameStringFromLocaleCode", "xspConvert result: %s\n", utf8str);
            utf8str = (unsigned char *)xspStrClone((char *)str_in_locale);
        } else {
            spDebug(10, "spSetID3TextFrameStringFromLocaleCode", "xspConvert failed\n");
        }
        size = 0;
    } else {
        long len_in_locale;
        spKanjiCode icode;
            
        len_in_locale = (long)strlen((char *)str_in_locale);

        size = (len_in_locale + 1) * 2;
        spDebug(80, "spSetID3TextFrameStringFromLocaleCode", "size = %ld, len_in_locale = %ld, str_in_locale = %s\n",
                size, len_in_locale, str_in_locale);
        utf8str = xspAlloc(size, unsigned char);
        spStrCopy((char *)utf8str, size, (char *)str_in_locale);
            
        /* locale code --> UTF8 */
        icode = spConvertKanjiFromLocaleCode(utf8str, size, SP_KANJI_CODE_UTF8);
        spDebug(80, "spSetID3TextFrameStringFromLocaleCode", "icode = %d, size = %ld\n", icode, size);
    }

    if (osize != NULL) *osize = size;

    return utf8str;
}

static spBool setID3TextFrameStringFromLocaleCode(spConverter conv_to_id3, spID3TextFrame *tframe, unsigned char encoding, const unsigned char *str_in_locale)
{
    long size = 0;
    unsigned char *ptr = NULL;
    unsigned char *utf8str = NULL;
    unsigned short *utf16str = NULL;
    unsigned char buf[SP_MAX_LINE];
    int genre_index = -1;
    
    if (tframe == NULL || str_in_locale == NULL) return SP_FALSE;

    spDebug(80, "spSetID3TextFrameStringFromLocaleCode", "encoding = %d, str_in_locale = %s, tframe->header.type = %s\n",
            encoding, str_in_locale, tframe->header.type);
    
    if (encoding == 0) {
        spID3Header *header;
        
        if ((header = spGetID3HeaderFromFrame((spID3Frame *)tframe)) == NULL) {
            return SP_FALSE;
        }
        
        if (getID3GenreIndexStringIfGenreFrame(header, tframe, str_in_locale, buf, sizeof(buf), &genre_index) == SP_TRUE) {
            ptr = buf;
            size = (long)strlen((char *)ptr);
        }
    }

    if (ptr == NULL) {
        if ((utf8str = xconvertLocaleCodeToUTF8String(conv_to_id3, str_in_locale, &size)) == NULL) {
            return SP_FALSE;
        }

        if (encoding == 1/* UTF-16 */ || encoding == 2/* UTF-16BE */) {
            if (encoding == 2) {
                tframe->bom_flag = SP_FALSE;
                tframe->swap_flag = SP_AIFF_NEED_SWAP;
            } else if (tframe->bom_flag == SP_FALSE) {
                /* need BOM */
                tframe->bom_flag = SP_TRUE;
                tframe->swap_flag = SP_WAV_NEED_SWAP; /* little endian such as iTunes */
            }
            size = MAX(size, (long)(strlen((char *)utf8str) + 1));
            spDebug(80, "spSetID3TextFrameStringFromLocaleCode", "UTF16 size = %ld\n", size);
            utf16str = xspAlloc(size, unsigned short);
        
            spUTF8ToUnicode16(utf8str, utf16str, size * 2);
            size = (long)spUnicode16Length(utf16str) * 2;
            ptr = (unsigned char *)utf16str;
        } else {
            ptr = utf8str;
            size = (long)strlen((char *)ptr);
        }
    }

    if (ptr == NULL) {
        ptr = (unsigned char *)str_in_locale;
        size = (long)strlen((char *)str_in_locale);
    }
    
    spDebug(80, "spSetID3TextFrameStringFromLocaleCode", "final string size = %ld\n", size);
    
    spSetID3TextFrameString(tframe, encoding, ptr, size);

    if (utf8str != NULL) {
        xspFree(utf8str);
    }
    if (utf16str != NULL) {
        xspFree(utf16str);
    }

    return SP_TRUE;
}

spBool spSetID3TextFrameStringFromLocaleCode(spID3TextFrame *tframe, unsigned char encoding, const unsigned char *str_in_locale)
{
    return setID3TextFrameStringFromLocaleCode(NULL, tframe, encoding, str_in_locale);
}

static spID3TextFrame *appendID3TextFrameStringFromLocaleCode(spConverter conv_to_id3, spID3Header *header, const char *type, spBool find_existing_frame, unsigned char encoding, const unsigned char *str_in_locale)
{
    spID3TextFrame *tframe = NULL;
    
    if (header == NULL || str_in_locale == NULL) return NULL;

    if (encoding != 0 && spIsStringInsideID3AsciiRange(str_in_locale) == SP_TRUE) {
        encoding = 0;
    }

    if (encoding == 3/* UTF-8 */ && header->version < 4) return NULL;
    
    if ((tframe = (spID3TextFrame *)spAppendID3Frame(header, type, find_existing_frame)) == NULL) {
        return NULL;
    }

    setID3TextFrameStringFromLocaleCode(conv_to_id3, tframe, encoding, str_in_locale);

    return tframe;
}

spID3TextFrame *spAppendID3TextFrameStringFromLocaleCode(spID3Header *header, const char *type, spBool find_existing_frame, unsigned char encoding, const unsigned char *str_in_locale)
{
    return appendID3TextFrameStringFromLocaleCode(NULL, header, type, find_existing_frame, encoding, str_in_locale);
}

spID3TextFrame *spAppendID3TextFrameString(spID3Header *header, const char *type, spBool find_existing_frame,
                                           unsigned char encoding, const unsigned char *str, long len)
{
    spID3TextFrame *tframe = NULL;
    
    if (header == NULL || (encoding == 3/* UTF-8 */ && header->version < 4) || str == NULL) return NULL;

    if ((tframe = (spID3TextFrame *)spAppendID3Frame(header, type, find_existing_frame)) == NULL) {
        return NULL;
    }
    
    spSetID3TextFrameString(tframe, encoding, str, len);

    return tframe;
}

static spID3TextFrame *appendID3TextFrameTrackOrDiscNumber(spID3Header *header, spLong32 number, spLong32 total, spBool disc_flag)
{
    char *type;
    char buf[SP_MAX_LINE];
    spID3TextFrame *tframe = NULL;

    if (disc_flag) {
        type = SP_ID3_FRAME_TYPE_DISC;
    } else {
        type = SP_ID3_FRAME_TYPE_TRACK;
    }
    
    if ((tframe = (spID3TextFrame *)spAppendID3Frame(header, type, SP_TRUE)) == NULL) {
        return NULL;
    }

    if (total > 0) {
        sprintf(buf, "%d/%d", (int)number, (int)total);
    } else {
        sprintf(buf, "%d", (int)number);
    }
    
    spSetID3TextFrameString(tframe, 0, (const unsigned char *)buf, (long)strlen(buf));

    return tframe;
}

spID3TextFrame *spAppendID3TextFrameTrackNumber(spID3Header *header, spLong32 number, spLong32 total)
{
    return appendID3TextFrameTrackOrDiscNumber(header, number, total, SP_FALSE);
}

spID3TextFrame *spAppendID3TextFrameDiscNumber(spID3Header *header, spLong32 number, spLong32 total)
{
    return appendID3TextFrameTrackOrDiscNumber(header, number, total, SP_TRUE);
}

spID3TextFrame *spAppendID3TextFrameTempo(spID3Header *header, spLong32 tempo)
{
    char buf[SP_MAX_LINE];
    spID3TextFrame *tframe = NULL;

    if ((tframe = (spID3TextFrame *)spAppendID3Frame(header, SP_ID3_FRAME_TYPE_TEMPO, SP_TRUE)) == NULL) {
        return NULL;
    }

    sprintf(buf, "%d", (int)tempo);
    
    spSetID3TextFrameString(tframe, 0, (unsigned char *)buf, (long)strlen(buf));

    return tframe;
}

static spBool parseID3SongInfo(spConverter conv_from_id3, spID3Header *header, spSongInfoV2 *song_info)
{
    if (header == NULL || song_info == NULL) return SP_FALSE;

    spInitSongInfoV2(song_info);

    if (header->version == 4) {
        song_info->info_mask |= SP_SONG_ID3V2_4_MASK;
    } else if (header->version == 2) {
        song_info->info_mask |= SP_SONG_ID3V2_2_MASK;
    } else {
        song_info->info_mask |= SP_SONG_ID3V2_3_MASK;
    }

    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_TITLE, (unsigned char *)song_info->title, sizeof(song_info->title))) {
        song_info->info_mask |= SP_SONG_TITLE_MASK;
    }
    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_ARTIST, (unsigned char *)song_info->artist, sizeof(song_info->artist))) {
        song_info->info_mask |= SP_SONG_ARTIST_MASK;
    }
    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_ALBUM, (unsigned char *)song_info->album, sizeof(song_info->album))) {
        song_info->info_mask |= SP_SONG_ALBUM_MASK;
    }
    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_GENRE, (unsigned char *)song_info->genre, sizeof(song_info->genre))) {
        song_info->info_mask |= SP_SONG_GENRE_MASK;
    }
    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_ALBUM_ARTIST, (unsigned char *)song_info->album_artist, sizeof(song_info->album_artist))) {
        song_info->info_mask |= SP_SONG_ALBUM_ARTIST_MASK;
    }
    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_RELEASE, (unsigned char *)song_info->release, sizeof(song_info->release))) {
        song_info->info_mask |= SP_SONG_RELEASE_MASK;
    }
    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_COPYRIGHT, (unsigned char *)song_info->copyright, sizeof(song_info->copyright))) {
        song_info->info_mask |= SP_SONG_COPYRIGHT_MASK;
    }
    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_ISRC, (unsigned char *)song_info->isrc, sizeof(song_info->isrc))) {
        song_info->info_mask |= SP_SONG_ISRC_MASK;
    }
    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_COMPOSER, (unsigned char *)song_info->composer, sizeof(song_info->composer))) {
        song_info->info_mask |= SP_SONG_COMPOSER_MASK;
    }
    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_LYRICIST, (unsigned char *)song_info->lyricist, sizeof(song_info->lyricist))) {
        song_info->info_mask |= SP_SONG_LYRICIST_MASK;
    }
    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_SOFTWARE, (unsigned char *)song_info->software, sizeof(song_info->software))) {
        song_info->info_mask |= SP_SONG_SOFTWARE_MASK;
    }
    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_PUBLISHER, (unsigned char *)song_info->source, sizeof(song_info->source))) {
        song_info->info_mask |= SP_SONG_SOURCE_MASK;
    }
    if (findID3TextFrameStringToLocaleCode(conv_from_id3, header, SP_ID3_FRAME_TYPE_COMMENT, (unsigned char *)song_info->comment, sizeof(song_info->comment))) {
        song_info->info_mask |= SP_SONG_COMMENT_MASK;
    }

    if (spFindID3TextFrameTrackNumber(header, &song_info->track, &song_info->track_total)) {
        if (song_info->track > 0) {
            song_info->info_mask |= SP_SONG_TRACK_MASK;
        }
        if (song_info->track_total > 0) {
            song_info->info_mask |= SP_SONG_TRACK_TOTAL_MASK;
        }
    }

    if (spFindID3TextFrameDiscNumber(header, &song_info->disc, &song_info->disc_total)) {
        if (song_info->disc > 0) {
            song_info->info_mask |= SP_SONG_DISC_MASK;
        }
        if (song_info->disc_total > 0) {
            song_info->info_mask |= SP_SONG_DISC_TOTAL_MASK;
        }
    }

    if (spFindID3TextFrameTempo(header, &song_info->tempo)) {
        song_info->info_mask |= SP_SONG_TEMPO_MASK;
    }
    
    if (song_info->info_mask) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spParseID3SongInfo(spID3Header *header, spSongInfoV2 *song_info)
{
    return parseID3SongInfo(NULL, header, song_info);
}

static spBool appendID3SongInfo(spConverter conv_to_id3, spID3Header *header, spSongInfoV2 *song_info)
{
    spLong32 number, total;
    
    if (header == NULL || song_info == NULL || song_info->info_mask == 0) return SP_FALSE;

    spDebug(10, "spAppendID3SongInfo", "song_info->info_mask = %lx\n", song_info->info_mask);

    if (song_info->info_mask & SP_SONG_ID3V2_4_MASK) {
        header->version = 4;
    } else if (song_info->info_mask & SP_SONG_ID3V2_3_MASK) {
        header->version = 3;
    } else if (song_info->info_mask & SP_SONG_ID3V2_2_MASK) {
        header->version = 2;
    } else {
        if (song_info->info_mask & SP_SONG_ID3V1_MASK) {
            return SP_FALSE;
        }
        /* default: ID3v2.3 */
        header->version = 3;
    }
    
    if (song_info->info_mask & SP_SONG_TITLE_MASK) {
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_TITLE, SP_TRUE, 1, (unsigned char *)song_info->title);
    }
    if (song_info->info_mask & SP_SONG_ARTIST_MASK) {
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_ARTIST, SP_TRUE, 1, (unsigned char *)song_info->artist);
    }
    if (song_info->info_mask & SP_SONG_ALBUM_ARTIST_MASK) {
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_ALBUM_ARTIST, SP_TRUE, 1, (unsigned char *)song_info->album_artist);
    }
    if (song_info->info_mask & SP_SONG_ALBUM_MASK) {
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_ALBUM, SP_TRUE, 1, (unsigned char *)song_info->album);
    }
    if (song_info->info_mask & SP_SONG_ISRC_MASK) {
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_ISRC, SP_TRUE, 1, (unsigned char *)song_info->isrc);
    }
    if (song_info->info_mask & SP_SONG_COMPOSER_MASK) {
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_COMPOSER, SP_TRUE, 1, (unsigned char *)song_info->composer);
    }
    if (song_info->info_mask & SP_SONG_LYRICIST_MASK) {
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_LYRICIST, SP_TRUE, 1, (unsigned char *)song_info->lyricist);
    }
    if (song_info->info_mask & SP_SONG_GENRE_MASK) {
        spDebug(80, "spAppendID3SongInfo", "call spAppendID3TextFrameStringFromLocaleCode: genre = %s\n", song_info->genre);
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_GENRE, SP_TRUE, 1, (unsigned char *)song_info->genre);
    }
    if (song_info->info_mask & SP_SONG_RELEASE_MASK) {
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_RELEASE, SP_TRUE, 1, (unsigned char *)song_info->release);
    }
    if (song_info->info_mask & SP_SONG_SOFTWARE_MASK) {
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_SOFTWARE, SP_TRUE, 1, (unsigned char *)song_info->software);
    }
    if (song_info->info_mask & SP_SONG_COPYRIGHT_MASK) {
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_COPYRIGHT, SP_TRUE, 1, (unsigned char *)song_info->copyright);
    }
    if (song_info->info_mask & SP_SONG_SOURCE_MASK) {
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_PUBLISHER, SP_TRUE, 1, (unsigned char *)song_info->source);
    }
    if (song_info->info_mask & SP_SONG_COMMENT_MASK) {
        appendID3TextFrameStringFromLocaleCode(conv_to_id3, header, SP_ID3_FRAME_TYPE_COMMENT, SP_TRUE, 1, (unsigned char *)song_info->comment);
    }

    number = total = 0;
    if (song_info->info_mask & SP_SONG_TRACK_MASK) {
        number = song_info->track;
        if (song_info->info_mask & SP_SONG_TRACK_TOTAL_MASK) {
            total = song_info->track_total;
        }
        spAppendID3TextFrameTrackNumber(header, number, total);
    }

    number = total = 0;
    if (song_info->info_mask & SP_SONG_DISC_MASK) {
        number = song_info->disc;
        if (song_info->info_mask & SP_SONG_DISC_TOTAL_MASK) {
            total = song_info->disc_total;
        }
        spAppendID3TextFrameDiscNumber(header, number, total);
    }

    if (song_info->info_mask & SP_SONG_TEMPO_MASK && song_info->tempo > 0) {
        spAppendID3TextFrameTempo(header, song_info->tempo);
    }

    return SP_TRUE;
}

spBool spAppendID3SongInfo(spID3Header *header, spSongInfoV2 *song_info)
{
    return appendID3SongInfo(NULL, header, song_info);
}

spBool spIsStringInsideID3AsciiRange(const unsigned char *str)
{
    int i;
    int len;

    len = (int)strlen((char *)str);

    for (i = 0; i < len; i++) {
        if (!((str[i] >= 0x20 && str[i] <= 0x7e) || str[i] == 0x0a)) {
            return SP_FALSE;
        }
    }

    return SP_TRUE;
}

spBool spIsStringInsideID3Laten1Range(const unsigned char *str)
{
    int i;
    int len;

    len = (int)strlen((char *)str);

    for (i = 0; i < len; i++) {
        if (!((str[i] >= 0x20 && str[i] <= 0x7e) || str[i] == 0x0a || (str[i] >= 0xa0 && str[i] <= 0xff))) {
            return SP_FALSE;
        }
    }

    return SP_TRUE;
}

static void showID3FrameString(spConverter conv_from_id3, long depth, char encoding, unsigned char *string, unsigned char *orig_string)
{
    long k;
    char buf[SP_MAX_LINE];
    
    spDebug(10, "showID3FrameString", "encoding = %d\n", encoding);
    
    if (encoding == 1 || encoding == 2) {
        spUnicode16ToUTF8((unsigned short *)string, (unsigned char *)buf, sizeof(buf));
    } else {
        spStrCopy(buf, sizeof(buf), (char *)string);
    }
            
    if (encoding != 0) {
        if (conv_from_id3 != NULL) {
            char *ostr;
            /* UTF8 --> system codeset */
            if ((ostr = xspConvert(conv_from_id3, buf)) != NULL) {
                spDebug(80, "showID3FrameString", "xspConvert result: %s\n", ostr);
                spStrCopy(buf, sizeof(buf), ostr);
                xspFree(ostr);
            } else {
                spDebug(10, "showID3FrameString", "xspConvert failed\n");
            }
        } else {
            spKanjiCode ocode;
            /* UTF8 --> locale code */
            ocode = spConvertKanjiToLocaleCode((unsigned char *)buf, sizeof(buf), SP_KANJI_CODE_UTF8);
            spDebug(80, "showID3FrameString", "ocode = %d\n", ocode);
        }
    }
    for (k = 1; k <= depth; k++) printf("  ");
    printf("String (encoding = %d): '%s'", encoding, buf);
    if (encoding == 0 && orig_string != NULL) {
        printf(" (Original: '%s')", orig_string);
    }
    printf("\n");
    
    return;
}

static void showID3FrameInfo(spConverter conv_from_id3, spID3Frame *frame, long depth)
{
    long k;
    char type_string[5];
    spID3Frame *child_frame;
    spID3FrameSize frame_size;
    unsigned char buf[SP_MAX_LINE];

    frame_size = spGetID3FrameSize(frame, type_string);
    spDebug(10, "showID3FrameInfo", "depth = %ld, frame_size = %ld\n", depth, (long)frame_size);

    if (frame_size > 0) {
        for (k = 1; k < depth; k++) printf("  ");
        printf("'%s' (%ld bytes)\n", type_string, (long)frame_size);

        if (strneq(type_string, "ID3", 3)) {
            spID3Header *header = (spID3Header *)frame;
            printf("  V2.%d.%d\n", header->version, header->revision);
        } else if (strneq(type_string, "COMM", 4) || strneq(type_string, "COM ", 4)
                   || strneq(type_string, "USLT", 4)|| strneq(type_string, "ULT ", 4)) {
            spID3LangTextFrame *ltframe = (spID3LangTextFrame *)frame;
            
            showID3FrameString(conv_from_id3, depth, ltframe->encoding, ltframe->string, NULL);
            if (ltframe->actual_string != NULL) {
                showID3FrameString(conv_from_id3, depth, ltframe->encoding, ltframe->actual_string, NULL);
            }
        } else if (strneq(type_string, "T", 1) || strneq(type_string, "COMM", 4)
            || strneq(type_string, "COM ", 4)) {
            int encoding;
            unsigned char *ptr;
            unsigned char *orig_string = NULL;
            int genre_index = -1;
            spID3Header *header;
            spID3TextFrame *tframe = (spID3TextFrame *)frame;

            encoding = tframe->encoding;
            ptr = tframe->string;
            
            if ((header = spGetID3HeaderFromFrame(frame)) != NULL) {
                if (getID3GenreStringIfGenreFrame(header, tframe, tframe->string, buf, sizeof(buf), &genre_index) == SP_TRUE
                    && genre_index >= 0) {
                    orig_string = tframe->string;
                    encoding = 0;
                    ptr = buf;
                }
            }

            showID3FrameString(conv_from_id3, depth, encoding, ptr, orig_string);
        }
    }
    
    child_frame = frame->child;
    while (child_frame != NULL) {
        showID3FrameInfo(conv_from_id3, child_frame, depth + 1);
        child_frame = child_frame->next;
    }

    return;
}

void spShowID3HeaderInfo(spID3Header *header)
{
    showID3FrameInfo(NULL, (spID3Frame *)header, 0);
    return;
}

static const char *sp_id3v1_genres_list[] = {
    "Blues",
    "Classic Rock",
    "Country",
    "Dance",
    "Disco",
    "Funk",
    "Grunge",
    "Hip-Hop",
    "Jazz",
    "Metal",
    "New Age",
    "Oldies",
    "Other",
    "Pop",
    "R&B",
    "Rap",
    "Reggae",
    "Rock",
    "Techno",
    "Industrial",
    "Alternative",
    "Ska",
    "Death Metal",
    "Pranks",
    "Soundtrack",
    "Euro-Techno",
    "Ambient",
    "Trip-Hop",
    "Vocal",
    "Jazz+Funk",
    "Fusion",
    "Trance",
    "Classical",
    "Instrumental",
    "Acid",
    "House",
    "Game",
    "Sound Clip",
    "Gospel",
    "Noise",
    "Alternative",
    "Bass",
    "Punk",
    "Space",
    "Meditative",
    "Instrumental Pop",
    "Instrumental Rock",
    "Ethnic",
    "Gothic",
    "Darkwave",
    "Techno-Industrial",
    "Electronic",
    "Pop-Folk",
    "Eurodance",
    "Dream",
    "Southern Rock",
    "Comedy",
    "Cult",
    "Gangsta",
    "Top 40",
    "Christian Rap",
    "Pop/Funk",
    "Jungle",
    "Native American",
    "Cabaret",
    "New Wave",
    "Psychedelic",
    "Rave",
    "Showtunes",
    "Trailer",
    "Lo-Fi",
    "Tribal",
    "Acid Punk",
    "Acid Jazz",
    "Polka",
    "Retro",
    "Musical",
    "Rock & Roll",
    "Hard Rock",
    "Folk",
    "Folk-Rock",
    "National Folk",
    "Swing",
    "Fast Fusion",
    "Bebob",
    "Latin",
    "Revival",
    "Celtic",
    "Bluegrass",
    "Avantgarde",
    "Gothic Rock",
    "Progressive Rock",
    "Psychedelic Rock",
    "Symphonic Rock",
    "Slow Rock",
    "Big Band",
    "Chorus",
    "Easy Listening",
    "Acoustic",
    "Humour",
    "Speech",
    "Chanson",
    "Opera",
    "Chamber Music",
    "Sonata",
    "Symphony",
    "Booty Bass",
    "Primus",
    "Porn Groove",
    "Satire",
    "Slow Jam",
    "Club",
    "Tango",
    "Samba",
    "Folklore",
    "Ballad",
    "Power Ballad",
    "Rhytmic Soul",
    "Freestyle",
    "Duet",
    "Punk",
    "Drum Solo",
    "Acapella",
    "Euro-House",
    "Dance Hall",
    "Goa",
    "Drum & Bass",
    "Club-House",
    "Hardcore",
    "Terror",
    "Indie",
    "BritPop",
    "Afro-punk",
    "Polsk Punk",
    "Beat",
    "Christian gangsta rap",
    "Heavy Metal",
    "Black Metal",
    "Crossover",
    "Contemporary Christian",
    "Christian Rock",
    "Merengue",
    "Salsa",
    "Thrash",
    "Anime",
    "JPop",
    "SynthPop",
    "Abstract",
    "Art Rock",
    "Baroque",
    "Bhangra",
    "Big Beat",
    "Breakbeat",
    "Chillout",
    "Downtempo",
    "Dub",
    "EBM",
    "Eclectic",
    "Electro",
    "Electroclash",
    "Emo",
    "Experimental",
    "Garage",
    "Global",
    "IDM",
    "Illbient",
    "Industro-Goth",
    "Jam Band",
    "Krautrock",
    "Leftfield",
    "Lounge",
    "Math Rock",
    "New Romantic",
    "Nu-Breakz",
    "Post-Punk",
    "Post-Rock",
    "Psytrance",
    "Shoegaze",
    "Space Rock",
    "Trop Rock",
    "World Music",
    "Neoclassical",
    "Audiobook",
    "Audio theatre",
    "Neue Deutsche Welle",
    "Podcast",
    "Indie-Rock",
    "G-Funk",
    "Dubstep",
    "Garage Rock",
    "Psybient",
    NULL,
};

const char *spGetID3v1GenreString(int index, spID3v1GenreExtension *extension_type)
{
    if (index < 0 || index >= 192) return NULL;

    if (extension_type != NULL) {
        if (index < SP_ID3V1_GENRE_EXTENSION_WINAMP_CORE) {
            *extension_type = SP_ID3V1_GENRE_NONEXTENSION;
        } else if (index < SP_ID3V1_GENRE_EXTENSION_WINAMP_1998) {
            *extension_type = SP_ID3V1_GENRE_EXTENSION_WINAMP_CORE;
        } else if (index < SP_ID3V1_GENRE_EXTENSION_WINAMP_2010) {
            *extension_type = SP_ID3V1_GENRE_EXTENSION_WINAMP_1998;
        } else {
            *extension_type = SP_ID3V1_GENRE_EXTENSION_WINAMP_2010;
        }
    }

    return sp_id3v1_genres_list[index];
}

int spFindID3v1GenreIndex(const char *string, spID3v1GenreExtension extension_limit)
{
    int i;
    int index = -1;

    if (strnone(string)) return -1;

    for (i = 0; sp_id3v1_genres_list[i] != NULL; i++) {
        if (((extension_limit == SP_ID3V1_GENRE_NONEXTENSION)
             && (i >= SP_ID3V1_GENRE_EXTENSION_WINAMP_CORE))
            || ((extension_limit == SP_ID3V1_GENRE_EXTENSION_WINAMP_CORE)
                && (i >= SP_ID3V1_GENRE_EXTENSION_WINAMP_1998))
            || ((extension_limit == SP_ID3V1_GENRE_EXTENSION_WINAMP_1998)
                && (i >= SP_ID3V1_GENRE_EXTENSION_WINAMP_2010))) {
            break;
        }

        if (strcaseeq(sp_id3v1_genres_list[i], string)) {
            index = i;
            break;
        }
    }

    return index;
}

#if 1
struct spID3Converter
{
    unsigned long options;
    char codeset[SP_MAX_LINE];
    spConverter conv_from_id3;
    spConverter conv_to_id3;
};

spID3Converter spID3ConverterOpen(const char *codeset, unsigned long options)
{
    spID3Converter id3conv;

    id3conv = xspAlloc(1, struct spID3Converter);
    memset(id3conv, 0, sizeof(struct spID3Converter));

    id3conv->options = options;
    spStrCopy(id3conv->codeset, sizeof(id3conv->codeset), codeset);

    return id3conv;
}

void spID3ConverterClose(spID3Converter id3conv)
{
    if (id3conv == NULL) return;
    
    if (id3conv->conv_from_id3 != NULL) {
        spCloseConverter(id3conv->conv_from_id3);
    }
    if (id3conv->conv_to_id3 != NULL) {
        spCloseConverter(id3conv->conv_to_id3);
    }
    xspFree(id3conv);
    
    return;
}

static spBool openStringConverterFromID3(spID3Converter id3conv)
{
    if ((id3conv->options & SP_ID3_CONVERTER_OPTION_NO_CODESET_CONVERTER)) {
        if (id3conv->conv_from_id3 != NULL) {
            spCloseConverter(id3conv->conv_from_id3);
            id3conv->conv_from_id3 = NULL;
        }
        return SP_TRUE;
    }
    
    if (strnone(id3conv->codeset) || id3conv->conv_from_id3 != NULL) return SP_TRUE;

    if ((id3conv->conv_from_id3 = spOpenConverter("UTF-8", id3conv->codeset)) != NULL) {
        spDebug(80, "openStringConverterFromID3", "spOpenConverter to %s succeeded\n", id3conv->codeset);
        return SP_TRUE;
    } else {
        spDebug(10, "openStringConverterFromID3", "spOpenConverter to %s failed\n", id3conv->codeset);
        return SP_FALSE;
    }
}

static spBool openStringConverterToID3(spID3Converter id3conv)
{
    if ((id3conv->options & SP_ID3_CONVERTER_OPTION_NO_CODESET_CONVERTER)) {
        if (id3conv->conv_to_id3 != NULL) {
            spCloseConverter(id3conv->conv_to_id3);
            id3conv->conv_to_id3 = NULL;
        }
        return SP_TRUE;
    }
    
    if (strnone(id3conv->codeset) || id3conv->conv_to_id3 != NULL) return SP_TRUE;

    if ((id3conv->conv_to_id3 = spOpenConverter(id3conv->codeset, "UTF-8")) != NULL) {
        spDebug(80, "openStringConverterToID3", "spOpenConverter from %s succeeded\n", id3conv->codeset);
        return SP_TRUE;
    } else {
        spDebug(10, "openStringConverterToID3", "spOpenConverter from %s failed\n", id3conv->codeset);
        return SP_FALSE;
    }
}

spBool spID3ConverterFindTextFrameString(spID3Converter id3conv, spID3Header *header, const char *type,
                                         unsigned char *str_in_codeset, long size)
{
    spID3TextFrame *tframe;

    if (id3conv == NULL) return SP_FALSE;

    if ((tframe = (spID3TextFrame *)spFindID3FrameFromHeader(header, type)) == NULL) {
        spDebug(10, "spID3ConverterFindTextFrameStringToLocaleCode", "cannot find frame: type = %s\n", type);
        return SP_FALSE;
    }

    if (id3conv->conv_from_id3 == NULL) {
        openStringConverterFromID3(id3conv);
    }
    
    return getID3TextFrameStringToLocaleCode(id3conv->conv_from_id3, tframe, str_in_codeset, size);
}

spID3TextFrame *spID3ConverterAppendTextFrameString(spID3Converter id3conv, spID3Header *header, const char *type,
                                                    spBool find_existing_frame, unsigned char encoding, const unsigned char *str_in_codeset)
{
    spID3TextFrame *tframe = NULL;
    
    if (id3conv == NULL || header == NULL || str_in_codeset == NULL) return NULL;

    if (encoding != 0 && spIsStringInsideID3AsciiRange(str_in_codeset) == SP_TRUE) {
        encoding = 0;
    }

    if (encoding == 3/* UTF-8 */ && header->version < 4) return NULL;
    
    if ((tframe = (spID3TextFrame *)spAppendID3Frame(header, type, find_existing_frame)) == NULL) {
        return NULL;
    }

    if (id3conv->conv_to_id3 == NULL) {
        openStringConverterToID3(id3conv);
    }
    
    setID3TextFrameStringFromLocaleCode(id3conv->conv_to_id3, tframe, encoding, str_in_codeset);

    return tframe;
}

spBool spID3ConverterParseSongInfo(spID3Converter id3conv, spID3Header *header, spSongInfoV2 *song_info)
{
    if (id3conv == NULL) return SP_FALSE;
    
    if (id3conv->conv_from_id3 == NULL) {
        openStringConverterFromID3(id3conv);
    }
    
    return parseID3SongInfo(id3conv->conv_from_id3, header, song_info);
}

spBool spID3ConverterAppendSongInfo(spID3Converter id3conv, spID3Header *header, spSongInfoV2 *song_info)
{
    if (id3conv == NULL) return SP_FALSE;

    if (id3conv->conv_to_id3 == NULL) {
        openStringConverterToID3(id3conv);
    }
    
    return appendID3SongInfo(id3conv->conv_to_id3, header, song_info);
}

void spID3ConverterShowID3HeaderInfo(spID3Converter id3conv, spID3Header *header)
{
    if (id3conv == NULL) return;

    if (id3conv->conv_from_id3 == NULL) {
        openStringConverterFromID3(id3conv);
    }
    
    showID3FrameInfo(id3conv->conv_from_id3, (spID3Frame *)header, 0);
    
    return;
}

spBool spID3ConverterConvertUTF8StringTo(spID3Converter id3conv, unsigned char *iobuf, long buf_size)
{
    if (id3conv == NULL) return SP_FALSE;

    if (id3conv->conv_from_id3 == NULL) {
        openStringConverterFromID3(id3conv);
    }

    convertUTF8StringToLocaleCode(id3conv->conv_from_id3, iobuf, buf_size);

    return SP_TRUE;
}

spBool spID3ConverterConvertToUTF8String(spID3Converter id3conv, unsigned char *iobuf, long buf_size)
{
    unsigned char *utf8str;
    
    if (id3conv == NULL) return SP_FALSE;

    if (id3conv->conv_to_id3 == NULL) {
        openStringConverterToID3(id3conv);
    }
    
    if ((utf8str = xconvertLocaleCodeToUTF8String(id3conv->conv_to_id3, iobuf, NULL)) != NULL) {
        spStrCopy((char *)iobuf, buf_size, (char *)utf8str);
        xspFree(utf8str);
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}
#endif
